# coding: utf-8

"""
    LMK API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.0.1
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import re  # noqa: F401

from pydantic import validate_arguments, ValidationError
from typing_extensions import Annotated

from pydantic import StrictStr

from lmk.generated.models.access_token_response import AccessTokenResponse
from lmk.generated.models.create_headless_auth_session_request import CreateHeadlessAuthSessionRequest
from lmk.generated.models.headless_auth_refresh_token_request import HeadlessAuthRefreshTokenRequest
from lmk.generated.models.headless_auth_session_response import HeadlessAuthSessionResponse

from lmk.generated.api_client import ApiClient
from lmk.generated.exceptions import (  # noqa: F401
    ApiTypeError,
    ApiValueError
)


class HeadlessAuthApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client

    @validate_arguments
    def create_headless_auth_session(self, create_headless_auth_session_request : CreateHeadlessAuthSessionRequest, **kwargs) -> HeadlessAuthSessionResponse:  # noqa: E501
        """  # noqa: E501

        Create a headless auth session, which allows certain LMK apps to retrieve oauth access tokens without any server component  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.create_headless_auth_session(create_headless_auth_session_request, async_req=True)
        >>> result = thread.get()

        :param create_headless_auth_session_request: (required)
        :type create_headless_auth_session_request: CreateHeadlessAuthSessionRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: HeadlessAuthSessionResponse
        """
        kwargs['_return_http_data_only'] = True
        return self.create_headless_auth_session_with_http_info(create_headless_auth_session_request, **kwargs)  # noqa: E501

    @validate_arguments
    def create_headless_auth_session_with_http_info(self, create_headless_auth_session_request : CreateHeadlessAuthSessionRequest, **kwargs):  # noqa: E501
        """  # noqa: E501

        Create a headless auth session, which allows certain LMK apps to retrieve oauth access tokens without any server component  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.create_headless_auth_session_with_http_info(create_headless_auth_session_request, async_req=True)
        >>> result = thread.get()

        :param create_headless_auth_session_request: (required)
        :type create_headless_auth_session_request: CreateHeadlessAuthSessionRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(HeadlessAuthSessionResponse, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = [
            'create_headless_auth_session_request'
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method create_headless_auth_session" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []

        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))

        # process the form parameters
        _form_params = []
        _files = {}

        # process the body parameter
        _body_params = None
        if _params['create_headless_auth_session_request']:
            _body_params = _params['create_headless_auth_session_request']

        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get('_content_type',
            self.api_client.select_header_content_type(
                ['application/json']))
        if _content_types_list:
                _header_params['Content-Type'] = _content_types_list

        # authentication setting
        _auth_settings = []  # noqa: E501

        _response_types_map = {
            '201': "HeadlessAuthSessionResponse",
        }

        return self.api_client.call_api(
            '/v1/headlessAuth', 'POST',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))

    @validate_arguments
    def get_headless_auth_session(self, session_id : StrictStr, **kwargs) -> HeadlessAuthSessionResponse:  # noqa: E501
        """  # noqa: E501

        Get the headless auth session with the given ID.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_headless_auth_session(session_id, async_req=True)
        >>> result = thread.get()

        :param session_id: (required)
        :type session_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: HeadlessAuthSessionResponse
        """
        kwargs['_return_http_data_only'] = True
        return self.get_headless_auth_session_with_http_info(session_id, **kwargs)  # noqa: E501

    @validate_arguments
    def get_headless_auth_session_with_http_info(self, session_id : StrictStr, **kwargs):  # noqa: E501
        """  # noqa: E501

        Get the headless auth session with the given ID.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.get_headless_auth_session_with_http_info(session_id, async_req=True)
        >>> result = thread.get()

        :param session_id: (required)
        :type session_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(HeadlessAuthSessionResponse, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = [
            'session_id'
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method get_headless_auth_session" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}
        if _params['session_id']:
            _path_params['sessionId'] = _params['session_id']

        # process the query parameters
        _query_params = []

        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))

        # process the form parameters
        _form_params = []
        _files = {}

        # process the body parameter
        _body_params = None

        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # authentication setting
        _auth_settings = []  # noqa: E501

        _response_types_map = {
            '200': "HeadlessAuthSessionResponse",
        }

        return self.api_client.call_api(
            '/v1/headlessAuth/{sessionId}', 'GET',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))

    @validate_arguments
    def refresh_headless_auth_token(self, headless_auth_refresh_token_request : HeadlessAuthRefreshTokenRequest, **kwargs) -> AccessTokenResponse:  # noqa: E501
        """  # noqa: E501

        Refresh an oauth access token obtained through a headless auth session  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.refresh_headless_auth_token(headless_auth_refresh_token_request, async_req=True)
        >>> result = thread.get()

        :param headless_auth_refresh_token_request: (required)
        :type headless_auth_refresh_token_request: HeadlessAuthRefreshTokenRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: AccessTokenResponse
        """
        kwargs['_return_http_data_only'] = True
        return self.refresh_headless_auth_token_with_http_info(headless_auth_refresh_token_request, **kwargs)  # noqa: E501

    @validate_arguments
    def refresh_headless_auth_token_with_http_info(self, headless_auth_refresh_token_request : HeadlessAuthRefreshTokenRequest, **kwargs):  # noqa: E501
        """  # noqa: E501

        Refresh an oauth access token obtained through a headless auth session  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.refresh_headless_auth_token_with_http_info(headless_auth_refresh_token_request, async_req=True)
        >>> result = thread.get()

        :param headless_auth_refresh_token_request: (required)
        :type headless_auth_refresh_token_request: HeadlessAuthRefreshTokenRequest
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(AccessTokenResponse, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = [
            'headless_auth_refresh_token_request'
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method refresh_headless_auth_token" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []

        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))

        # process the form parameters
        _form_params = []
        _files = {}

        # process the body parameter
        _body_params = None
        if _params['headless_auth_refresh_token_request']:
            _body_params = _params['headless_auth_refresh_token_request']

        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get('_content_type',
            self.api_client.select_header_content_type(
                ['application/json']))
        if _content_types_list:
                _header_params['Content-Type'] = _content_types_list

        # authentication setting
        _auth_settings = []  # noqa: E501

        _response_types_map = {
            '201': "AccessTokenResponse",
        }

        return self.api_client.call_api(
            '/v1/headlessAuth/refresh', 'POST',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))

    @validate_arguments
    def retrieve_headless_auth_session_token(self, session_id : StrictStr, **kwargs) -> AccessTokenResponse:  # noqa: E501
        """  # noqa: E501

        Retrieve the oauth token for the headless auth session. If the session is complete, the access and refresh tokens will be returned. Otherwise, this endpoint will return an error code.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.retrieve_headless_auth_session_token(session_id, async_req=True)
        >>> result = thread.get()

        :param session_id: (required)
        :type session_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: AccessTokenResponse
        """
        kwargs['_return_http_data_only'] = True
        return self.retrieve_headless_auth_session_token_with_http_info(session_id, **kwargs)  # noqa: E501

    @validate_arguments
    def retrieve_headless_auth_session_token_with_http_info(self, session_id : StrictStr, **kwargs):  # noqa: E501
        """  # noqa: E501

        Retrieve the oauth token for the headless auth session. If the session is complete, the access and refresh tokens will be returned. Otherwise, this endpoint will return an error code.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.retrieve_headless_auth_session_token_with_http_info(session_id, async_req=True)
        >>> result = thread.get()

        :param session_id: (required)
        :type session_id: str
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(AccessTokenResponse, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = [
            'session_id'
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method retrieve_headless_auth_session_token" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}
        if _params['session_id']:
            _path_params['sessionId'] = _params['session_id']

        # process the query parameters
        _query_params = []

        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))

        # process the form parameters
        _form_params = []
        _files = {}

        # process the body parameter
        _body_params = None

        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # authentication setting
        _auth_settings = []  # noqa: E501

        _response_types_map = {
            '200': "AccessTokenResponse",
        }

        return self.api_client.call_api(
            '/v1/headlessAuth/{sessionId}/token', 'GET',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))



import asyncio
import os
import signal

from lmk.processmon.attach import attach_simple
from lmk.processmon.child_monitor import ChildMonitor
from lmk.processmon.daemon import ProcessMonitorController, pid_ctx
from lmk.processmon.manager import JobManager
from lmk.utils import async_signal_handler_ctx, asyncio_create_task, setup_logging


async def main() -> None:
    setup_logging()

    base_path = os.path.expanduser("~/.lmk")
    manager = JobManager(base_path)

    monitor = ChildMonitor(["python", "test.py"])

    job = manager.create_job("python")
    print("JOB ID", job.job_id)

    with pid_ctx(job.pid_file, os.getpid()):
        controller = ProcessMonitorController(
            -1,
            monitor,
            job.job_dir,
        )
        run_task = asyncio_create_task(controller.run())
        attach_task = asyncio_create_task(attach_simple(job.job_dir))

        async with async_signal_handler_ctx(
            [signal.SIGINT, signal.SIGTERM],
            lambda signum: controller.send_signal(signum)
        ):
            await asyncio.wait([run_task, attach_task])


if __name__ == "__main__":
    asyncio.run(main())



# coding: utf-8

"""
    LMK API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.0.1
    Generated by: https://openapi-generator.tech
"""


from __future__ import annotations
from inspect import getfullargspec
import pprint
import re  # noqa: F401
import json


from typing import Optional
from pydantic import BaseModel, Field, StrictStr, validator

class JupyterSessionState(BaseModel):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    type: StrictStr = ...
    url: Optional[StrictStr] = None
    notebook_name: Optional[StrictStr] = Field(None, alias="notebookName")
    shell_state: Optional[StrictStr] = Field(None, alias="shellState")
    cell_state: Optional[StrictStr] = Field(None, alias="cellState")
    cell_text: Optional[StrictStr] = Field(None, alias="cellText")
    cell_error: Optional[StrictStr] = Field(None, alias="cellError")
    execution_num: Optional[float] = Field(None, alias="executionNum")
    cell_started_at: Optional[StrictStr] = Field(None, alias="cellStartedAt")
    cell_finished_at: Optional[StrictStr] = Field(None, alias="cellFinishedAt")
    notify_on: StrictStr = Field(..., alias="notifyOn")
    notify_channel: Optional[StrictStr] = Field(None, alias="notifyChannel")
    __properties = ["type", "url", "notebookName", "shellState", "cellState", "cellText", "cellError", "executionNum", "cellStartedAt", "cellFinishedAt", "notifyOn", "notifyChannel"]

    @validator('type')
    def type_validate_enum(cls, v):
        if v not in ('jupyter'):
            raise ValueError("must validate the enum values ('jupyter')")
        return v

    class Config:
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> JupyterSessionState:
        """Create an instance of JupyterSessionState from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> JupyterSessionState:
        """Create an instance of JupyterSessionState from a dict"""
        if obj is None:
            return None

        if type(obj) is not dict:
            return JupyterSessionState.parse_obj(obj)

        _obj = JupyterSessionState.parse_obj({
            "type": obj.get("type"),
            "url": obj.get("url"),
            "notebook_name": obj.get("notebookName"),
            "shell_state": obj.get("shellState"),
            "cell_state": obj.get("cellState"),
            "cell_text": obj.get("cellText"),
            "cell_error": obj.get("cellError"),
            "execution_num": obj.get("executionNum"),
            "cell_started_at": obj.get("cellStartedAt"),
            "cell_finished_at": obj.get("cellFinishedAt"),
            "notify_on": obj.get("notifyOn"),
            "notify_channel": obj.get("notifyChannel")
        })
        return _obj




import contextlib
import logging
from typing import Generator


def is_jupyter() -> bool:
    try:
        from IPython import get_ipython
    except ImportError:
        return False
    shell = get_ipython()
    if shell is None:
        return False
    shell_class_name = type(shell).__name__
    return shell_class_name == "ZMQInteractiveShell"


def run_javascript(js: str) -> None:
    if not is_jupyter():
        raise RuntimeError("Cannot run JS outside of a Jupyter Notebook context")

    from IPython.display import Javascript, display

    display(Javascript(js))


def display_widget() -> None:
    """
    Imported as _ipython_display_ in lmk/jupyter/__init__.py
    to act as a display hook for the module
    """
    try:
        from IPython.display import display
        from lmk.jupyter.widget import get_widget
    except ImportError as err:
        # TODO: improve
        raise RuntimeError(f"Jupyter modules not installed") from err
    widget = get_widget()
    display(widget)


@contextlib.contextmanager
def background_ctx(logger: logging.Logger, ctx: str) -> Generator[None, None, None]:
    try:
        yield
    except Exception as err:
        logger.exception(f"Error encountered in {ctx}")



import shlex
from typing import List


def shlex_join(args: List[str]) -> str:
    # This is only available in python 3.8+
    if hasattr(shlex, "join"):
        return shlex.join(args)

    return " ".join(shlex.quote(arg) for arg in args)



from datetime import datetime
from typing import List, Optional

from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column


class Base(DeclarativeBase):
    pass


class Job(Base):
    """ """

    __tablename__ = "job"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(unique=True)
    pid: Mapped[Optional[int]]
    command: Mapped[Optional[str]]
    notify_on: Mapped[str]
    notify_status: Mapped[Optional[str]]
    error_type: Mapped[Optional[str]]
    error: Mapped[Optional[str]]
    channel_id: Mapped[Optional[str]]
    session_id: Mapped[Optional[str]]
    started_at: Mapped[Optional[datetime]]
    ended_at: Mapped[Optional[datetime]]
    exit_code: Mapped[Optional[int]]

    def is_running(self) -> bool:
        if self.ended_at:
            return False
        if not self.started_at:
            return False
        return True



import click
import os
import psutil
import shlex
import signal as signal_module
import sys
import textwrap
from typing import List, Optional, Dict, Any

from lmk.instance import get_instance, set_instance, Instance
from lmk.process import exc
from lmk.process.attach import attach_interactive
from lmk.process.child_monitor import ChildMonitor
from lmk.process.client import send_signal, update_job
from lmk.process.lldb_monitor import LLDBProcessMonitor, check_lldb
from lmk.process.logging import get_log_level
from lmk.process.manager import JobManager
from lmk.process.run import run_foreground, run_daemon
from lmk.process.shell_plugin import (
    detect_shell,
    install_script,
    uninstall_script,
    shell_profile_file,
    resolve_pid,
    get_shell_cli_script,
)
from lmk.utils.click import async_command, async_group
from lmk.utils.decorators import stack_decorators
from lmk.utils.logging import setup_logging


# Sigh, just to make the vercel build work; sqlite3
# docs not exist there, so the cli() setup fails. Side
# note though--is that the wrong place to be doing that setup
# if it runs _even on a --help command???_
DOCS_ONLY = bool(os.getenv("LMK_CLI_DOCS_ONLY"))


def _check_login() -> None:
    instance = get_instance()
    if instance.logged_in():
        return

    resp = click.confirm("You are not logged in; would you like to log in now?")
    if not resp:
        raise exc.NotLoggedIn()

    instance.login(force=True)


cli_args = stack_decorators(
    click.option(
        "-l", "--log-level", default="WARN", help="Log level, defaults to WARN"
    ),
    click.option(
        "-b",
        "--base-path",
        default=None,
        help="Path to the LMK configuration directory; defaults to ~/.lmk",
    ),
)


@async_group(
    help=(
        "The LMK Command Line Interface. This allows you to monitor command-line processes remotely "
        "via the LMK web app."
    )
)
@cli_args
@click.pass_context
async def cli(ctx: click.Context, log_level: str, base_path: str):
    if DOCS_ONLY:
        return

    log_level = get_log_level(log_level)

    setup_logging(level=log_level)
    ctx.ensure_object(dict)
    ctx.obj["log_level"] = log_level

    manager = JobManager(base_path)
    await manager.setup()

    ctx.obj["manager"] = manager

    if base_path is not None:
        config_path = os.path.join(base_path, "config")
        set_instance(Instance(config_path=config_path))


@cli.command(
    help=(
        "Authenticate with the LMK API. This should typically be the first command you run when "
        "using the CLI for the first time, and it allows you to use the other commands."
    )
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help=(
        "By default, this will be a no-op if you are already logged in. Pass --force to "
        "force the CLI to re-authenticate with LMK."
    ),
)
@click.option(
    "-m",
    "--manual",
    is_flag=True,
    default=False,
    help="Print the authentication URL instead of attempting to open the page in a web browser.",
)
def login(force, manual):
    instance = get_instance()
    instance.login(force=force, auth_mode="manual" if manual else None)


attach_option = click.option(
    "--attach/--no-attach",
    default=True,
    help=(
        "Attach to the process, meaning you will see the logs from the process appearing in your "
        "console in real time. You can detach at any time, at which point you can choose to let the "
        "process keep running or interrupt it. After detaching, you will also be able to re-attach at any "
        "time while the process is running."
    ),
)

name_option = click.option(
    "-N",
    "--name",
    default=None,
    help=(
        "Job name for the job you want to run. This will be visible in the LMK dashboard, "
        "and it can be passed to other commands that take a job ID such as `attach`, `kill`, etc. "
        "If not passed, a name will be generated."
    ),
)


def notify_on_option(default: str = "none"):
    return click.option(
        "-n",
        "--notify",
        default=default,
        type=click.Choice(["stop", "error", "none"]),
        help=(
            "Set the initial `notify_on` value for the process. If `stop`, you will get a notification "
            "whenever the process completes for any reason. If `error`, you will only get a notification "
            "if it exists with a non-zero status code. If `none`, you will note be notified at all, but will "
            "still be able to monitor the process of the script via the LMK web app. This can be changed after "
            "launching the process initially, so if you omit it initially you can always choose to notify yourself "
            "later."
        ),
    )


@async_command(
    cli,
    short_help="Run a command line process, and monitor it using LMK",
    help=(
        "Run a command line process, and monitor it using LMK. This means that you will be able "
        "to notify yourself when it finishes or finishes with an error, and interrupt it remotely if "
        "you choose."
    ),
)
@click.option(
    "--daemon/--no-daemon",
    default=True,
    help=(
        "Daemonize the monitored process. This is the default, and it means that you can detach "
        "and re-attach to the process while it's running without interrupting it."
    ),
)
@attach_option
@name_option
@notify_on_option()
@click.argument("command", nargs=-1)
@click.pass_context
async def run(
    ctx: click.Context,
    daemon: bool,
    command: List[str],
    attach: bool,
    name: Optional[str],
    notify: str,
):
    if not command:
        click.secho("No command provided", fg="red")
        raise click.Abort

    head, *tail = command
    command = shlex.split(head) + tail

    if name is None:
        name = command[0]

    _check_login()

    manager: JobManager = ctx.obj["manager"]
    job = await manager.create_job(name, notify)
    click.secho(f"Job ID: {job.name}", fg="green", bold=True)

    monitor = ChildMonitor(command)

    if daemon:
        await run_daemon(job.name, monitor, manager, ctx.obj["log_level"])
        if attach:
            exit_code = await attach_interactive(job.name, manager)
            sys.exit(exit_code or 0)
    else:
        await run_foreground(job.name, monitor, manager, ctx.obj["log_level"], attach)


monitor_args = stack_decorators(
    click.argument("pid"),
    attach_option,
    name_option,
    notify_on_option(),
    click.option("-j", "--job", default=None, hidden=True),
)


@async_command(
    cli,
    short_help="Monitor a script that has already been started",
    help=(
        "Monitor a script that has already been started and either paused via Ctrl-Z or "
        "is running in a different terminal."
    ),
)
@monitor_args
@click.pass_context
async def monitor(
    ctx: click.Context,
    pid: str,
    attach: bool,
    name: Optional[str],
    notify: str,
    job: str,
):
    resolved_pid, _ = resolve_pid(pid)

    manager: JobManager = ctx.obj["manager"]
    if name is None:
        proc = psutil.Process(resolved_pid)
        name = proc.cmdline()[0]

    _check_login()

    await check_lldb()

    if job is None:
        job_obj = await manager.create_job(name, notify)
    else:
        job_obj = await manager.get_job(job, not_started=True)
        if job_obj is None:
            raise exc.JobNotFound(job)
        if notify != job_obj.notify_on:
            await manager.update_job(job, notify_on=notify)

    click.secho(f"Job ID: {job_obj.name}", fg="green", bold=True)

    monitor = LLDBProcessMonitor(resolved_pid)

    await run_daemon(job_obj.name, monitor, manager, ctx.obj["log_level"])

    if attach:
        exit_code = await attach_interactive(job_obj.name, manager)
        sys.exit(exit_code or 0)


@async_command(
    cli,
    short_help="Attach to a monitored job",
    help=(
        "Attach to a monitored job, so you will see the output in your terminal in real time. You can "
        "detach from or interrupt the job while attached."
    ),
)
@click.argument("job_id")
@click.pass_context
async def attach(ctx: click.Context, job_id: str):
    manager: JobManager = ctx.obj["manager"]
    job = await manager.get_job(job_id)

    if job is None:
        raise exc.JobNotFound(job_id)

    if not job.is_running():
        raise exc.JobNotRunning(job_id)

    exit_code = await attach_interactive(job.name, manager)
    sys.exit(exit_code or 0)


def pad(value: str, length: int, character: str = " ", **style_kwargs) -> str:
    if len(value) > length:
        return value[: length - 3] + "..."
    return click.style(value + character * (length - len(value)), **style_kwargs)


@async_command(cli, help="List jobs monitored by LMK")
@click.option(
    "-a",
    "--all",
    is_flag=True,
    help="List all jobs; by default only running jobs are shown",
)
@click.pass_context
async def jobs(ctx: click.Context, all: bool):
    manager: JobManager = ctx.obj["manager"]
    jobs = await manager.list_jobs(running_only=not all)

    if not jobs:
        click.echo("No jobs found")
        return

    click.echo(
        " ".join(
            [
                pad("name", 30, bold=True),
                pad("pid", 10, bold=True),
                pad("status", 12, bold=True),
                pad("notify", 10, bold=True),
                pad("started", 30, bold=True),
            ]
        )
    )
    for job in jobs:
        state_kwargs: Dict[str, Any] = {}
        if job.ended_at:
            exit_str = job.exit_code if job.exit_code is not None else "?"
            if job.exit_code == 0:
                state_kwargs = {"fg": "yellow"}
            else:
                state_kwargs = {"fg": "red"}
            state = f"exited ({exit_str})"
        elif job.started_at:
            state = "running"
            state_kwargs = {"fg": "green", "bold": True}
        else:
            state = "not-started"

        notify_on = ""
        notify_on_kwargs = {}
        if job.notify_on == "stop":
            notify_on = "stop"
            notify_on_kwargs = {"fg": "yellow"}
        elif job.notify_on == "error":
            notify_on = "error"
            notify_on_kwargs = {"fg": "red"}

        click.echo(
            " ".join(
                [
                    pad(job.name, 30, bold=True),
                    pad(str(job.pid), 10),
                    pad(state, 12, **state_kwargs),
                    pad(notify_on, 10, **notify_on_kwargs),
                    pad(job.started_at.isoformat() if job.started_at else "", 30),
                ]
            )
        )


@async_command(
    cli,
    short_help="Send a signal to a monitored job",
    help="Send a signal to a monitored job to interrupt or terminate it.",
)
@click.argument("job_id")
@click.option(
    "-s", "--signal", default="SIGINT", help="Signal to send, defaults to SIGINT"
)
@click.pass_context
async def kill(ctx: click.Context, job_id: str, signal: str):
    manager: JobManager = ctx.obj["manager"]
    job = await manager.get_job(job_id)

    if job is None:
        raise exc.JobNotFound(job_id)

    if not job.is_running():
        raise exc.JobNotRunning(job_id)

    signal_value: Optional[int]
    if signal.isdigit():
        signal_value = int(signal)
    else:
        signal_value = getattr(
            signal_module,
            signal.upper(),
            getattr(signal_module, "SIG" + signal.upper(), None),
        )

    if signal_value is None:
        click.secho(f"Invalid signal value: {signal}", fg="red", bold=True)
        raise click.Abort

    await send_signal(manager.socket_file(job.name), signal_value)


@async_command(
    cli,
    short_help="Change the `notify_on` value for a running job",
    help=(
        "Change the `notify_on` value for a running job to a new value. This will set the job to "
        "notify you when it's finished running, exits with an error, or change it to not notify you "
        "if it's currently configured to send you a notification."
    ),
)
@click.argument("job_id")
@notify_on_option("stop")
@click.pass_context
async def notify(ctx: click.Context, job_id: str, notify: str) -> None:
    manager: JobManager = ctx.obj["manager"]
    job = await manager.get_job(job_id)

    if job is None:
        raise exc.JobNotFound(job_id)

    if not job.is_running():
        raise exc.JobNotRunning(job_id)

    await manager.update_job(job.name, notify_on=notify)
    await update_job(manager.socket_file(job.name))


@cli.command(
    short_help="Install the LMK shell plugin",
    help=(
        "Install the LMK shell plugin, which allows you to monitor jobs "
        "using shell syntax like %1, %2, etc."
    ),
)
@click.option(
    "-i",
    "--install",
    default=False,
    is_flag=True,
    help="Install the shell CLI script in the shell profile",
)
@click.option(
    "-u",
    "--uninstall",
    default=False,
    is_flag=True,
    help="Uninstall the shell CLI script from the shell profile",
)
@click.option(
    "-p",
    "--print",
    is_flag=True,
    default=False,
    help=(
        "Print the shell plugin script; this is useful if the automatic installation process "
        "via --install is not working properly."
    ),
)
@click.option(
    "-s",
    "--shell",
    default=None,
    help=(
        "Shell flavor that you're using, e.g. bash or zsh. If this is not passed, "
        "the shell you're currently using will be detected automatically."
    ),
)
@click.pass_context
def shell_plugin(
    ctx: click.Context,
    install: bool,
    uninstall: bool,
    shell: Optional[str],
    print: bool,
) -> None:
    if shell is None:
        shell = detect_shell()

    num_actions = sum([install, uninstall, print])

    if num_actions > 1:
        ctx.fail("Only one of --print, --install, or --uninstall may be provided")
    if num_actions == 0:
        ctx.fail("One of --print, --install, or --uninstall must be provided")

    if print:
        script = get_shell_cli_script(shell)
        click.echo(script)
        click.echo(
            textwrap.dedent(
                """
        # THIS SHOULD BE SOURCED BY YOUR SHELL
        # E.g. . <(lmk shell-plugin --print)
        # To add to your shell profile, use the --install flag
        """
            ).strip()
        )
        return

    profile_file = shell_profile_file(shell)

    if install:
        install_script(profile_file)
        click.secho(
            f"Shell plugin installed successfully. Run "
            f"`. {profile_file}` or open a new shell to see the "
            f"changes take effect.",
            fg="green",
            bold=True,
        )
        return

    if uninstall:
        uninstall_script(profile_file)
        click.secho("Shell plugin uninstalled successfully", fg="green", bold=True)


@async_command(
    cli,
    short_help="Check if you can monitor already-running scripts on your system",
    help=(
        "Check if monitoring a process after it's already started is "
        "supported on your computer. See the docs for details: "
        "https://docs.lmkapp.dev/docs/cli/running-process."
    ),
)
async def check_existing_script_monitoring() -> None:
    try:
        await check_lldb()
    except exc.LLDBNotFound:
        click.secho(
            "\n".join([
                click.style(
                    "No supported debugger found, your system is not "
                    "supported. See the docs for details on setting up "
                    "your system: ",
                    fg="red"
                ),
                click.style(
                    "https://docs.lmkapp.dev/docs/cli/running-process",
                    bold=True
                )
            ])
        )
        sys.exit(1)
    except exc.LLDBCannotAttach:
        click.secho(
            "\n".join([
                click.style(
                    "Unable to attach to a process with lldb, your system is not "
                    "supported. See the docs for details on setting up "
                    "your system: ",
                    fg="red"
                ),
                click.style(
                    "https://docs.lmkapp.dev/docs/cli/running-process",
                    bold=True
                )
            ])
        )
        sys.exit(1)



import asyncio
import psutil
import sys
from typing import List, Any, Dict

import click

from lmk.cli import cli
from lmk.process.manager import JobManager
from lmk.process.shell_plugin import resolve_pid
from lmk.utils import shlex_join


def format_params(values: Dict[str, Any], params: List[click.Parameter]) -> List[str]:
    params_by_name = {param.name: param for param in params}

    out: List[str] = []
    for key, value in values.items():
        if key not in params_by_name:
            continue

        param = params_by_name[key]
        if value == param.default:
            continue

        if isinstance(param, click.Argument):
            if isinstance(value, tuple):
                out.extend(value)
            else:
                out.append(value)
        elif param.type == click.BOOL:
            out.append(param.opts[0] if value else param.secondary_opts[0])
        else:
            out.append(param.opts[-1])
            out.append(value)

    return out


async def process_cmd(
    ctx: click.Context,
    root_args: List[str],
) -> None:
    manager = JobManager(ctx.params["base_path"])

    await manager.setup()

    all_args = [*ctx.protected_args, *ctx.args]
    cmd_name, cmd, cmd_args = cli.resolve_command(ctx, all_args)
    with cmd.make_context(cmd_name, cmd_args, parent=ctx) as sub_ctx:
        cmd_params = cmd.get_params(sub_ctx)
        if cmd_name == "monitor":
            pid, shell_job_id = resolve_pid(sub_ctx.params["pid"])
            name = sub_ctx.params["name"]
            if name is None:
                proc = psutil.Process(pid)
                name = proc.cmdline()[0]

            if sub_ctx.params["job"] is None:
                job = await manager.create_job(name)
                job_id = job.name
            else:
                job_id = sub_ctx.params["job"]

            new_params = {
                **sub_ctx.params,
                "job": job_id,
                "pid": str(pid),
                "attach": False,
            }
            new_args = root_args.copy()
            new_args.append(cmd_name)
            new_args.extend(format_params(new_params, cmd_params))
            print(
                "CMD" if sub_ctx.params["attach"] else "LASTCMD", shlex_join(new_args)
            )

            print("DISOWN", shell_job_id)

            if sub_ctx.params["attach"]:
                new_args = root_args.copy()
                new_args.append("attach")
                attach_command = cli.get_command(ctx, "attach")
                attach_params = attach_command.get_params(ctx)
                new_attach_params = {"job_id": job_id}
                new_args.extend(format_params(new_attach_params, attach_params))
                print("LASTCMD", shlex_join(new_args))
        else:
            out_args = root_args.copy()
            out_args.append(cmd_name)
            out_args.extend(format_params(sub_ctx.params, cmd_params))
            print("LASTCMD", shlex_join(out_args))


async def main(args: List[str]) -> int:
    try:
        with cli.make_context("lmk", args=args) as ctx:
            out_args = format_params(ctx.params, cli.get_params(ctx))
            await process_cmd(ctx, out_args)
            return 0
    except click.exceptions.Abort:
        click.echo("Aborted!", file=sys.stderr)
        return 1
    except click.exceptions.ClickException as err:
        err.show()
        return err.exit_code
    except click.exceptions.Exit as err:
        return err.exit_code


if __name__ == "__main__":
    sys.exit(asyncio.run(main(sys.argv[1:])))



# Original: https://opensource.apple.com/source/lldb/lldb-179.1/examples/python/process_events.py
import argparse
import asyncio
import enum
import json
import logging
import os
import signal
import sys
from typing import List, Optional, Any

import lldb


LOGGER = logging.getLogger("lldb_monitor_script")


def setup_logging(level: str) -> None:
    LOGGER.setLevel(
        getattr(logging, level.strip().upper(), logging.INFO)
    )
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(
        logging.Formatter("%(asctime)s [%(name)s - %(levelname)s] %(message)s")
    )
    LOGGER.addHandler(handler)


class ProcessState(enum.Enum):
    """
    """
    Attaching = lldb.eStateAttaching
    Connected = lldb.eStateConnected
    Crashed = lldb.eStateCrashed
    Detached = lldb.eStateDetached
    Exited = lldb.eStateExited
    Invalid = lldb.eStateInvalid
    Launching = lldb.eStateLaunching
    Running = lldb.eStateRunning
    Stepping = lldb.eStateStepping
    Stopped = lldb.eStateStopped
    Suspended = lldb.eStateSuspended
    Unloaded = lldb.eStateUnloaded


class ThreadStopReason(enum.Enum):
    """
    """
    Exec = lldb.eStopReasonExec
    Breakpoint = lldb.eStopReasonBreakpoint
    Exception = lldb.eStopReasonException
    Instrumentation = lldb.eStopReasonInstrumentation
    Invalid = lldb.eStopReasonInvalid
    None_ = lldb.eStopReasonNone
    PlanComplete = lldb.eStopReasonPlanComplete
    Signal = lldb.eStopReasonSignal
    ThreadExiting = lldb.eStopReasonThreadExiting
    Trace = lldb.eStopReasonTrace


def run_commands(command_interpreter, commands):
    return_obj = lldb.SBCommandReturnObject()
    for command in commands:
        command_interpreter.HandleCommand( command, return_obj )
        if return_obj.Succeeded():
            LOGGER.debug("Command `%s` was successful:\n%s", command, return_obj.GetOutput())
        else:
            LOGGER.error("Command `%s` failed:\n%s", command, return_obj.GetError())


async def wait_for_event(listener: lldb.SBListener, timeout: float) -> Optional[lldb.SBEvent]:
    """
    """
    event = lldb.SBEvent()
    loop = asyncio.get_running_loop()
    result = await loop.run_in_executor(
        None,
        lambda: listener.WaitForEvent(timeout, event)
    )
    return event if result else None


async def wait_for_fd(fd: int) -> None:
    loop = asyncio.get_running_loop()
    future = asyncio.Future()
    loop.add_reader(fd, future.set_result, None)
    future.add_done_callback(lambda f: loop.remove_reader(fd))
    await future


def process_message(process: lldb.SBProcess, message: Any) -> None:
    if message["type"] == "send_signal":
        process.Signal(message["signal"])
        return

    LOGGER.warn("Unhandled message type: %s", message["type"])


async def run(process: lldb.SBProcess, listener: lldb.SBListener) -> int:
    pid = process.GetProcessID()

    LOGGER.debug("Running w/ pid: %s, process: %s, state: %s", pid, process, ProcessState(process.GetState()))

    stdin_fd = sys.stdin.fileno()

    timeout = 1
    stdin_task = asyncio.create_task(wait_for_fd(stdin_fd))
    event_task = asyncio.create_task(wait_for_event(listener, timeout))

    while True:
        await asyncio.wait([stdin_task, event_task], return_when=asyncio.FIRST_COMPLETED)
        if stdin_task.done():
            line = os.read(stdin_fd, 1000)
            try:
                message = json.loads(line)
                LOGGER.debug("New message: %s", message)
                process_message(process, message)
            except json.JSONDecodeError:
                LOGGER.exception("Error processing message")
            stdin_task = asyncio.create_task(wait_for_fd(stdin_fd))

        if event_task.done():
            event = event_task.result()
            event_task = asyncio.create_task(wait_for_event(listener, timeout))

            if not event:
                continue
            
            stream = lldb.SBStream()
            event.GetDescription(stream)

            state_int = lldb.SBProcess.GetStateFromEvent(event)
            state = ProcessState(state_int)
            LOGGER.debug("State: %s, %s", state, stream.GetData())

            if state == ProcessState.Stopped:
                thread = process.GetSelectedThread()
                stop_reason = ThreadStopReason(thread.GetStopReason())
                print("THREAD", stop_reason)
                if stop_reason == ThreadStopReason.Signal:
                    sig = signal.Signals(thread.GetStopReasonDataAtIndex(0))
                    print("SIGNAL", sig)
                LOGGER.info("process %d stopped", pid)
                continue

            if state == ProcessState.Exited:
                exit_desc = process.GetExitDescription()
                if exit_desc:
                    LOGGER.info("process %d exited with status %s: %s", pid, process.GetExitStatus(), exit_desc)
                else:
                    LOGGER.info("process %d exited with status %s", pid, process.GetExitStatus())
                return process.GetExitStatus()
            
            if state == ProcessState.Crashed:
                LOGGER.info("process %d crashed", pid)
                exit_status = -1
                try:
                    exit_status = process.GetExitStatus()
                except Exception:
                    LOGGER.exception("Error getting exit status")
                return exit_status
            
            if state == ProcessState.Detached:
                LOGGER.error("process %d detached unexpectedly", pid)
                return -1

            if state == ProcessState.Running:
                LOGGER.info("process %d resumed", pid)
                continue

            if state == ProcessState.Unloaded:
                LOGGER.info("process %d unloaded, this shouldn't happen", pid)
                continue

            if state == ProcessState.Connected:
                LOGGER.info("process %d connected", pid)
                continue

            if state == ProcessState.Attaching:
                LOGGER.info("process %d attaching", pid)
                continue

            if state == ProcessState.Launching:
                LOGGER.info("process %d launching")
                continue

            LOGGER.warn("Unhandled state: %s", state)


async def main(argv: List[str]) -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("pid", type=int, help="Pid to monitor")
    parser.add_argument("output_file", help="File to redirect output to")
    parser.add_argument("-l", "--log-level", default="INFO", help="Set log level")

    args = parser.parse_args(argv)

    setup_logging(args.log_level)
    
    LOGGER.info("Stopping %d", args.pid)
    os.kill(args.pid, signal.SIGSTOP)

    debugger = lldb.SBDebugger.Create(False)
    debugger.SetAsync(True)

    command_interpreter = debugger.GetCommandInterpreter()

    signals = {
        sig.name for sig in signal.Signals
    } - {"SIGKILL"}
    stop_signals = {"SIGTSTP", "SIGSTOP"}
    signals -= stop_signals

    run_commands(
        command_interpreter,
        [
            f"process handle -p true -s false -n true {' '.join(signals)}",
            f"process handle -p true -s true -n true {' '.join(stop_signals)}"
        ]
    )

    target = debugger.CreateTarget("")

    if not target:
        LOGGER.error("Target launch failed")
        return 1

    error = lldb.SBError()
    listener = lldb.SBListener("event_listener")

    process = target.AttachToProcessWithID(listener, args.pid,  error)
    if not process:
        LOGGER.error("Process attach failed")
        return 1

    if process.GetProcessID() == lldb.LLDB_INVALID_PROCESS_ID:
        LOGGER.error("Invalid pid: %d", args.pid)
        return 1

    if not error.Success():
        stream = lldb.SBStream()
        error.GetDescription(stream)
        LOGGER.error("Process attach failed: %s", stream.GetData())
        return 1

    detached_msg = (
        f"\n{args.pid}: This process is now being monitored by LMK\n"
        f"Output is now being redirected to {args.output_file}\n"
        f"Run \\`disown\\` so it will stay running even if you "
        f"close your terminal"
    )

    run_commands(
        command_interpreter,
        [
            f"expr int $fd = (int) open({json.dumps(args.output_file)}, 1089)",
            f"expr int $xd = (int) dup(1)",
            f"expr (void) dup2($fd, 1)",
            f"expr (void) dup2($fd, 2)",
            f"expr (void) write($xd, {json.dumps(detached_msg)}, {len(detached_msg)})",
            f"expr (void) close($fd)",
            f"expr (void) close($xd)",
        ]
    )
    process.Continue()
    # thread = process.GetSelectedThread()
    # print("STOPPED", thread.IsStopped())
    # print("SUSPENDED", thread.IsSuspended())
    # thread.StepInstruction(False)
    # print("STOPPED", thread.IsStopped())
    # print("SUSPENDED", thread.IsSuspended())
    # print("")
    # print("INFO", process.GetProcessInfo())
    # print("ALIVE", process.__get_is_alive__())
    # print("RUNNING", process.__get_is_running__())
    # print("STOPPED", process.__get_is_stopped__())
    try:
        exit_code = await run(process, listener)
        sys.stdout.write(json.dumps({"exit_code": exit_code}) + "\n")
        sys.stdout.flush()
    finally:
        LOGGER.info("Terminating...")
        process.Detach()
        debugger.Terminate()


if __name__ == "__main__":
    asyncio.run(main(sys.argv[1:]))



import asyncio
import os

from lmk.processmon.manager import JobManager


def pad(value: str, length: int, character: str = " ") -> str:
    if len(value) > length:
        return value[:length - 3] + "..."
    return value + character * (length - len(value))


async def main():
    base_path = os.path.expanduser("~/.lmk")
    manager = JobManager(base_path)
    print(
        pad("id", 30),
        pad("pid", 10),
        pad("status", 12),
        pad("notify", 10),
        pad("started", 30),
    )
    # job_ids = manager.get_all_job_ids()
    # async for job in manager.get_jobs(job_ids):
    async for job in manager.list_running_jobs():
        print(
            pad(str(job.job_id), 30),
            pad(str(job.target_pid), 10),
            pad("running    " if job.running else "not-running", 12),
            pad(job.notify_on, 10),
            pad(job.started_at.isoformat(), 30),
        )


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import configparser
import contextlib
import enum
import inspect
import json
import logging
import os
import threading
import time
import webbrowser
from datetime import datetime, timedelta
from typing import (
    Optional,
    List,
    Any,
    Dict,
    AsyncGenerator,
    Union,
    Awaitable,
    cast,
    ContextManager,
    AsyncContextManager,
)

import aiohttp
from blinker import signal
from dateutil.parser import parse as parse_dt

from lmk import exc
from lmk.api_client import api_client
from lmk.constants import APP_ID, API_URL
from lmk.generated.api.event_api import EventApi
from lmk.generated.api.headless_auth_api import HeadlessAuthApi
from lmk.generated.api.notification_api import NotificationApi
from lmk.generated.api.session_api import SessionApi
from lmk.generated.exceptions import ApiException
from lmk.generated.models.access_token_response import AccessTokenResponse
from lmk.generated.models.create_headless_auth_session_request import (
    CreateHeadlessAuthSessionRequest,
)
from lmk.generated.models.create_session_request import CreateSessionRequest
from lmk.generated.models.event_request import (
    EventRequest,
    EventNotificationConfiguration,
)
from lmk.generated.models.event_response import EventResponse
from lmk.generated.models.headless_auth_refresh_token_request import (
    HeadlessAuthRefreshTokenRequest,
)
from lmk.generated.models.headless_auth_session_response import (
    HeadlessAuthSessionResponse,
)
from lmk.generated.models.notification_channel_response import (
    NotificationChannelResponse,
)
from lmk.generated.models.create_session_request_state import CreateSessionRequestState
from lmk.generated.models.process_session_state import ProcessSessionState
from lmk.generated.models.jupyter_session_state import JupyterSessionState
from lmk.generated.models.session_response import SessionResponse
from lmk.jupyter import is_jupyter, run_javascript
from lmk.utils.asyncio import async_callback, asyncio_lock
from lmk.utils.ws import WebSocket, ws_connected


LOGGER = logging.getLogger(__name__)

default_instance_changed = signal("default-instance-changed")

access_token_changed = signal("access-token-changed")

server_url_changed = signal("server-url-changed")

default_channel_changed = signal("default-channel-changed")

channels_fetch_state_changed = signal("channels-fetch-state-changed")


class ChannelType(str, enum.Enum):
    """ """

    Email = "email"
    TextMessage = "text-message"


class ChannelsState(str, enum.Enum):
    None_ = "none"
    Loading = "loading"
    Forbidden = "forbidden"
    Loaded = "loaded"
    Error = "error"


def is_context_manager(obj: Any) -> bool:
    return hasattr(obj, "__enter__") and hasattr(obj, "__exit__")


def is_async_context_manager(obj: Any) -> bool:
    return hasattr(obj, "__aenter__") and hasattr(obj, "__aexit__")


def pipeline(is_async: bool = False) -> Any:
    """
    Basic abstraction to be able to write sync & async implementations
    of the same API methods with the same code--provide a sequence of functions
    that return either values or coroutines depending on is_async, and this
    will resolve them in order, passing each result to the next function
    in the pipeline
    """

    def sync_pipeline(*funcs, ctx_managers: List[Any] = []):
        with contextlib.ExitStack() as stack:
            for ctx in ctx_managers:
                stack.enter_context(ctx)

            last_result = None
            for func in funcs:
                last_result = func(last_result)
            return last_result

    async def async_pipeline(*funcs, ctx_managers: List[Any] = []):
        async with contextlib.AsyncExitStack() as stack:
            for ctx in ctx_managers:
                if is_async_context_manager(ctx):
                    await stack.enter_async_context(ctx)
                else:
                    stack.enter_context(ctx)

            last_result = None
            for func in funcs:
                last_result = func(last_result)
                if inspect.isawaitable(last_result):
                    last_result = await last_result
            return last_result

    return async_pipeline if is_async else sync_pipeline


def handle_error(is_async: bool = False) -> Any:
    """
    Similar to pipeline(), but for handling errors
    """

    def sync_handle_error(func, error_type, handle_error):
        try:
            return func()
        except error_type as err:
            return handle_error(err)

    async def async_handle_error(func, error_type, handle_error):
        try:
            return await func()
        except error_type as err:
            return handle_error(err)

    return async_handle_error if is_async else sync_handle_error


class Channels:
    """
    An instance of this class is available at the top level of the ``lmk`` module under the
    ``lmk.channels`` attribute. This class's function is to simplify fetching and searching
    for notification channels. Channels are always fetched lazily when required based on
    ``get()``, ``list()``, or iterating through the ``channels`` object (synchronously or
    asynchronously).

    <details><summary>Usage Example</summary>
    <p>

    ```python
    import lmk

    # Check that the client is logged in (optional)
    # If using for the first time on a new device, call lmk.login()
    assert lmk.logged_in()

    # Get an email notification channel
    channel = lmk.channels.get(type="email")
    # Get a text message notification channel asynchronously
    channel = await lmk.channels.get(type="text-message", async_req=True)

    # Iterate through notification channels
    for channel in lmk.channels:
        print(channel)

    # Iterate through notification channels asynchronously
    async for channel in lmk.channels:
        print(channel)
    ```
    </p>
    </details>
    """

    def __init__(self, instance: "Instance") -> None:
        self.instance = instance
        self._fetch_state = ChannelsState.None_
        self._fetch_lock = threading.Lock()
        self._afetch_lock = asyncio_lock()
        self.data: Optional[List[NotificationChannelResponse]] = None

        @access_token_changed.connect_via(instance)
        def handle_access_token_changed(sender, old_value, new_value):
            if not new_value:
                self.fetch_state = ChannelsState.None_
                self.data = None

    @property
    def fetch_state(self) -> ChannelsState:
        return self._fetch_state

    @fetch_state.setter
    def fetch_state(self, value: Union[ChannelsState, str]) -> None:
        if isinstance(value, str) and not isinstance(value, ChannelsState):
            value = ChannelsState(value)
        if value == self._fetch_state:
            return
        old_value, self._fetch_state = self._fetch_state, value
        channels_fetch_state_changed.send(self, old_value=old_value, new_value=value)

    def _ensure_fetched(self, fetch: bool = False, async_req: bool = False) -> None:
        if self.fetch_state != ChannelsState.Loaded and not fetch:
            raise Exception

        def maybe_fetch():
            if self.fetch_state != ChannelsState.Loaded and fetch:
                return self.fetch(async_req=async_req)
            return None

        def check():
            if self.fetch_state != ChannelsState.Loaded:
                raise exc.ChannelsNotFetched(self.fetch_state)

        return pipeline(async_req)(lambda _: maybe_fetch(), lambda _: check())

    @property
    def default(self) -> Optional[str]:
        """
        The default notification channel; If ``notify()`` is called without passing
        ``notification_channels``, the notification will be sent to this channel. If you
        do not set this value, it will be set to the default notification channel for
        your account.

        <details><summary>Usage Example</summary>
        <p>

        ```python
        import lmk

        lmk.channels.default = lmk.channels.get(type="text-message")
        ```

        </p>
        </details>
        """
        return self.instance.default_channel

    @default.setter
    def default(self, value: Optional[Union[str, NotificationChannelResponse]]) -> None:
        self.instance.default_channel = value  # type: ignore

    def fetch(self, async_req: bool = False, force: bool = False) -> None:
        """
        Fetch notification channels.

        **Note:** This requires the client to be logged in.

        :param async_req: ``True`` if you want to send the request asynchronously, in which case this
        method will return a coroutine. Defaults to ``False``.
        :type async_req: bool, optional
        :param force: If ``True``, refetch notification channels even if they have already
        been fetched successfully. By default, fetching will be skipped if it's already been done
        successfully.

        :return: This method does not return anything
        :rtype: None
        """
        ctx_managers: List[Union[ContextManager[Any], AsyncContextManager[Any]]] = []
        if async_req:
            ctx_managers.append(self._afetch_lock)
        ctx_managers.append(self._fetch_lock)

        if (
            self.fetch_state
            not in {ChannelsState.None_, ChannelsState.Error, ChannelsState.Forbidden}
            and not force
        ):
            return

        self.fetch_state = ChannelsState.Loading

        def handle_channels(channels: List[NotificationChannelResponse]):
            LOGGER.debug("Channels: %s", channels)
            self.data = channels
            if channels:
                self.default = channels[0].notification_channel_id
            else:
                self.default = None
            self.fetch_state = ChannelsState.Loaded

        def handle_error_value(error: Exception):
            if isinstance(error, ApiException) and error.status == 403:
                self.fetch_state = ChannelsState.Forbidden
            else:
                self.fetch_state = ChannelsState.Error
            raise error

        error_handler = handle_error(async_req)

        return error_handler(
            lambda: pipeline(async_req)(
                lambda _: self.instance.list_notification_channels(async_req),
                handle_channels,
                ctx_managers=ctx_managers,
            ),
            Exception,
            handle_error_value,
        )

    def __repr__(self) -> str:
        if self.fetch_state != ChannelsState.Loaded:
            return f"{type(self).__name__}(fetch_state={self.fetch_state})"
        if not self.data:
            return f"{type(self).__name__}(<no channels>)"

        channel_strs = []
        for channel in self.data:
            is_default = channel.notification_channel_id == self.default
            prefix = "*" if is_default else " "
            channel_strs.append(
                " ".join(
                    [
                        " ",
                        prefix,
                        channel.name,
                        f"({channel.payload.actual_instance.type})",
                    ]
                )
            )

        channels_str = "\n".join(channel_strs)

        return f"{type(self).__name__}(\n{channels_str}\n)"

    def __iter__(self):
        self._ensure_fetched(fetch=True)
        yield from self.data

    async def __aiter__(self):
        await self._ensure_fetched(fetch=True, async_req=True)
        for channel in self.data:
            yield channel

    def list(
        self,
        name: Optional[str] = None,
        type: Optional[Union[str, ChannelType]] = None,
        name_exact: bool = False,
        fetch: bool = True,
        async_req: bool = False,
    ) -> List[NotificationChannelResponse]:
        """
        List notification channels. This will fetch notification channels if they haven't been
        fetched yet.

        :param name: Filter down to notification channels with the given name
        :type name: str, optional
        :param name_exact: Only return channels whose names exactly match ``name``. By default,
        the name matching is case insensitive.
        :type name_exact: bool, optional
        :param type: Filter to only notification channels of this type
        :type type: str | ChannelType, optional
        :param fetch: Fetch notification channels if they haven't been fetched yet.
        :type fetch: bool, optional
        :param async_req: ``True`` if you want to send the request asynchronously, in which case this
        method will return a coroutine. Defaults to ``False``.
        :type async_req: bool, optional

        :return: A list of notification channels matching the given parameters
        :rtype: List[NotificationChannelResponse]
        """
        if isinstance(type, str) and not isinstance(type, ChannelType):
            type = ChannelType(type)

        def filter():
            channels = self.data
            if type is not None:
                channels = [
                    c for c in channels if c.payload.actual_instance.type == type.value
                ]
            if name is not None and name_exact:
                channels = [c for c in channels if c.name == name]
            if name is not None and not name_exact:
                channels = [c for c in channels if name in c.name.lower()]
            return channels

        return pipeline(async_req)(
            lambda _: self._ensure_fetched(fetch=fetch, async_req=async_req),
            lambda _: filter(),
        )

    def get(
        self,
        name: Optional[str] = None,
        type: Optional[Union[str, ChannelType]] = None,
        name_exact: bool = False,
        fetch: bool = True,
        async_req: bool = False,
    ) -> Optional[NotificationChannelResponse]:
        """
        Get a single notification channel with the given parameters.

        :param name: Filter down to notification channels with the given name
        :type name: str, optional
        :param name_exact: Only return channels whose names exactly match ``name``. By default,
        the name matching is case insensitive.
        :type name_exact: bool, optional
        :param type: Filter to only notification channels of this type
        :type type: str | ChannelType, optional
        :param fetch: Fetch notification channels if they haven't been fetched yet.
        :type fetch: bool, optional
        :param async_req: ``True`` if you want to send the request asynchronously, in which case this
        method will return a coroutine. Defaults to ``False``.
        :type async_req: bool, optional

        :return: A notification channel matching the given parameters, or ``None`` if none exists.
        :rtype: NotificationChannelResponse | None
        """

        def pick(channels):
            if len(channels) > 1:
                raise exc.MultipleChannelsMatched(channels)
            if len(channels) == 0:
                return None
            return channels[0]

        return pipeline(async_req)(
            lambda _: self.list(
                name=name,
                type=type,
                name_exact=name_exact,
                fetch=fetch,
                async_req=async_req,
            ),
            pick,
        )


class Instance:
    """ """

    access_token_expires: Optional[int]

    def __init__(
        self,
        profile: Optional[str] = None,
        server_url: Optional[str] = None,
        config_path: Optional[str] = None,
        access_token: Optional[str] = None,
        refresh_token: Optional[str] = None,
        access_token_expires: Optional[Union[datetime, str]] = None,
        logger: Optional[logging.Logger] = None,
        sync_config: bool = True,
    ) -> None:
        if profile is None:
            profile = os.getenv("LMK_PROFILE")
        if profile is None:
            profile = "python"
        if config_path is None:
            config_path = os.getenv("LMK_CONFIG_PATH")
        if access_token is None:
            access_token = os.getenv("LMK_ACCESS_TOKEN")
        if refresh_token is None:
            refresh_token = os.getenv("LMK_REFRESH_TOKEN")
        if access_token_expires is None:
            access_token_expires = os.getenv("LMK_ACCESS_TOKEN_EXPIRES")
        if server_url is None:
            server_url = os.getenv("LMK_SERVER_URL")
        if logger is None:
            logger = LOGGER

        access_token_expires_value: Optional[int]
        if isinstance(access_token_expires, str) and access_token_expires.isdigit():
            access_token_expires_value = int(access_token_expires)
        elif isinstance(access_token_expires, str):
            access_token_expires = parse_dt(access_token_expires)

        if isinstance(access_token_expires, datetime):
            access_token_expires_value = int(access_token_expires.timestamp() * 1000)
        if access_token_expires is None:
            access_token_expires_value = None

        self._config_loaded: bool = False
        self._access_token: Optional[str] = None
        self._server_url: Optional[str] = None
        self._default_channel: Optional[str] = None

        self.client = api_client(
            server_url=server_url or API_URL,
            logger=logger,
        )
        self.channels = Channels(self)

        # Set this to False before loading initial values so that we
        # don't overwrite things.
        self.sync_config = False

        self.profile = profile
        self.config_path = config_path
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.access_token_expires = access_token_expires_value
        self.server_url = cast(str, server_url)
        self.sync_config = sync_config

        self._load_config()

    def close(self) -> None:
        self.client.close()

    @property
    def access_token(self) -> Optional[str]:
        return self._access_token

    @access_token.setter
    def access_token(self, value: Optional[str]) -> None:
        if value == self.access_token:
            return
        old_value, self._access_token = self._access_token, value
        access_token_changed.send(self, old_value=old_value, new_value=value)

    @property
    def server_url(self) -> str:
        return self._server_url or API_URL

    @server_url.setter
    def server_url(self, value: Optional[str]) -> None:
        if value == self._server_url:
            return
        old_value, self._server_url = self._server_url, value

        self.client.configuration.host = value
        server_url_changed.send(
            self,
            old_value=old_value,
            new_value=value,
        )
        if self.sync_config:
            self._save_config()

    @property
    def default_channel(self) -> Optional[str]:
        return self._default_channel

    @default_channel.setter
    def default_channel(
        self, value: Optional[Union[str, NotificationChannelResponse]]
    ) -> None:
        LOGGER.debug(
            "Setting default channel; Current: %s, new: %s", self.default_channel, value
        )
        if isinstance(value, NotificationChannelResponse):
            value = value.notification_channel_id
        if value == self._default_channel:
            return
        old_value, self._default_channel = self._default_channel, value
        default_channel_changed.send(
            self,
            old_value=old_value,
            new_value=value,
        )

    def _load_config(self, force: bool = False, overwrite: bool = False) -> None:
        if self._config_loaded and not force:
            return

        parser = configparser.ConfigParser()
        if self.config_path is None:
            config_paths = [os.path.expanduser("~/.lmk/config")]
        elif not os.path.isfile(self.config_path):
            raise exc.ConfigFileNotFound(self.config_path)
        else:
            config_paths = [self.config_path]

        read_config_paths = [path for path in config_paths if os.path.isfile(path)]

        parser.read(read_config_paths)
        if self.profile in parser:
            section = parser[self.profile]
            if self.access_token is None or overwrite:
                self.access_token = section.get("access_token", self.access_token)
            if self.refresh_token is None or overwrite:
                self.refresh_token = section.get("refresh_token", self.refresh_token)
            if self.access_token_expires is None or overwrite:
                self.access_token_expires = section.getint(
                    "access_token_expires", self.access_token_expires
                )
            if self._server_url is None or overwrite:
                self.server_url = section.get("server_url", self._server_url)

        self._config_loaded = True

    def _save_config(self) -> None:
        config_path = self.config_path
        if config_path is None:
            config_path = os.path.expanduser("~/.lmk/config")

        config_dir = os.path.dirname(config_path)
        if not os.path.exists(config_dir):
            os.makedirs(config_dir)

        parser = configparser.ConfigParser()
        obj: Dict[str, Any] = {}
        if self.access_token is not None:
            obj["access_token"] = self.access_token
        if self.refresh_token is not None:
            obj["refresh_token"] = self.refresh_token
        if self.access_token_expires is not None:
            obj["access_token_expires"] = self.access_token_expires
        if self._server_url is not None:
            obj["server_url"] = self.server_url

        parser[self.profile] = obj
        with open(config_path, "w+") as f:
            parser.write(f)

        LOGGER.info("wrote config to %s", config_path)

        self._config_loaded = True

    def _refresh_access_token_sync(self) -> None:
        api = HeadlessAuthApi(self.client)
        response = api.refresh_headless_auth_token(
            HeadlessAuthRefreshTokenRequest(
                appId=APP_ID, refreshToken=cast(str, self.refresh_token)
            )
        )
        self.set_access_token(
            response.access_token,
            self.refresh_token,
            datetime.utcnow() + timedelta(seconds=response.expires_in),
        )

    def _get_access_token_sync(self) -> str:
        now = datetime.utcnow().timestamp() * 1000
        if self.access_token_expires and self.access_token_expires < now:
            self._refresh_access_token_sync()
        if self.access_token is None:
            raise exc.NotLoggedIn()

        return self.access_token

    async def _refresh_access_token_async(self) -> None:
        api = HeadlessAuthApi(self.client)
        response = await api.refresh_headless_auth_token(  # type: ignore
            HeadlessAuthRefreshTokenRequest(
                appId=APP_ID, refreshToken=cast(str, self.refresh_token)
            ),
            async_req=True,
        )
        self.set_access_token(
            response.access_token,
            self.refresh_token,
            datetime.utcnow() + timedelta(seconds=response.expires_in),
        )

    async def _get_access_token_async(self) -> str:
        now = datetime.utcnow().timestamp() * 1000
        if self.access_token_expires and self.access_token_expires < now:
            await self._refresh_access_token_async()
        if self.access_token is None:
            raise exc.NotLoggedIn()

        return self.access_token

    def _get_access_token(self, async_req: bool = False) -> Union[str, Awaitable[str]]:
        if async_req:
            return self._get_access_token_async()
        return self._get_access_token_sync()

    def logged_in(self) -> bool:
        """
        Check whether the client is currently logged in

        :return: a boolean indicating if you are logged in or not
        :rtype: bool
        """
        return bool(self.access_token)

    def logout(self) -> None:
        """
        Get rid of the current access token. After calling this, you will
        need to log in again in order to use methods that require authentication
        such as ``notify()``

        :return: This method does not return anything
        :rtype: None
        """
        self.access_token = None
        self.refresh_token = None
        self.access_token_expires = None
        if self.sync_config:
            self._save_config()

    def initiate_auth(
        self, scope: Optional[str] = None, async_req: bool = False
    ) -> HeadlessAuthSessionResponse:
        if scope is None:
            scope = "event.publish event.notify channel.read session.create"

        api = HeadlessAuthApi(self.client)
        return api.create_headless_auth_session(
            CreateHeadlessAuthSessionRequest(appId=APP_ID, scope=scope),
            async_req=async_req,
        )

    def retrieve_auth_token(
        self, session_id: str, async_req: bool = False
    ) -> AccessTokenResponse:
        api = HeadlessAuthApi(self.client)
        handler = handle_error(async_req)

        def handle_api_exception(err):
            if err.status == 410:
                raise exc.AccessTokenAlreadyRetrieved(session_id) from err
            if err.status == 412:
                raise exc.AuthSessionNotComplete(session_id)
            raise err

        return handler(
            lambda: api.retrieve_headless_auth_session_token(
                session_id=session_id,
                async_req=async_req,
            ),
            ApiException,
            handle_api_exception,
        )

    def set_access_token(
        self,
        access_token: str,
        refresh_token: Optional[str] = None,
        access_token_expires: Optional[datetime] = None,
    ) -> None:
        self.access_token = access_token
        self.refresh_token = refresh_token
        if isinstance(access_token_expires, datetime):
            self.access_token_expires = int(access_token_expires.timestamp() * 1000)
        else:
            self.access_token_expires = access_token_expires
        if self.sync_config:
            self._save_config()

    def login(
        self,
        scope: Optional[str] = None,
        timeout: float = 300.0,
        poll_interval: float = 1.0,
        auth_mode: Optional[str] = None,
        force: bool = False,
    ) -> None:
        """
        Authenticate LMK interactively. If possible, this will open
        a new auth session in a web browser. If not, it will print the
        auth URL and you will have to navigate to it in a browser manually

        :param auth_mode: ``jupyter``, ``browser`` or ``manual``. If ``None``, the first of those
        three that is available in the current context will be chosen automatically.
        :type auth_mode: str, optional
        :param scope: If desired, you may manually specify scopes that the retrieved access token
        should have, separated by spaces. If ``None``, all scopes will be requested. The user may
        modify these scopes during the OAuth flow to make them more restrictive.
        :type scope: str, optional
        :param timeout: A float indicating how long the client should wait for authentication to
        succeed before considering the authentication attempt failed.
        :type timeout: float, optional
        :param poll_interval: The poll interval for checking if the created authentication
        session has succeeded.
        :type poll_interval: float, optional
        :param force: By default if you are already logged in, this method will simply
        return immediately. By passing ``force=True``, you may force the client to replace
        the current authentication information and log in again.
        :type force: bool, optional

        :return: This method does not return anything
        :rtype: None
        """
        if self.logged_in() and not force:
            print("Already authenticated, pass force=True to re-authenticate")
            return

        session = self.initiate_auth(scope)

        if auth_mode is None and is_jupyter():
            auth_mode = "jupyter"

        if auth_mode is None:
            try:
                # This will raise if none is available
                webbrowser.get()
                auth_mode = "browser"
            except webbrowser.Error:
                LOGGER.debug("No web browser available")

        if auth_mode is None:
            auth_mode = "manual"

        if auth_mode == "jupyter":
            run_javascript(f"window.open({json.dumps(session.authorize_url)})")
        elif auth_mode == "browser":
            webbrowser.open(session.authorize_url)
        elif auth_mode == "manual":
            print(
                f"Authenticate in a web browser by navigating to {session.authorize_url}"
            )
        else:
            raise RuntimeError(
                f"Invalid auth_mode: {auth_mode}, options: manual, browser, jupyter"
            )

        start = time.time()
        while time.time() - start < timeout:
            try:
                response = self.retrieve_auth_token(session.session_id)
            except exc.AuthSessionNotComplete:
                time.sleep(poll_interval)
                continue
            else:
                self.set_access_token(
                    response.access_token,
                    response.refresh_token,
                    datetime.utcnow() + timedelta(seconds=response.expires_in),
                )
                break

        print("Authentication successful")

    def list_notification_channels(
        self, async_req: bool = False
    ) -> List[NotificationChannelResponse]:
        """
        List the notification channels that are available to send notifications to

        :param async_req: ``True`` if you want to send the request asynchronously, in which case this
        method will return a coroutine. Defaults to ``False``.
        :type async_req: bool, optional

        :return: A list of notification channels that the current client has access to.
        :rtype: List[NotificationChannelResponse]
        """
        api = NotificationApi(self.client)
        return pipeline(async_req)(
            lambda _: self._get_access_token(async_req),
            lambda access_token: api.list_notification_channels(
                async_req=async_req,
                _headers={"Authorization": f"Bearer {access_token}"},
            ),
            lambda response: response.channels,
        )

    def notify(
        self,
        message: str,
        content_type: str = "text/markdown",
        notification_channels: Optional[
            List[Union[str, NotificationChannelResponse]]
        ] = None,
        notify: bool = True,
        async_req: bool = False,
    ) -> EventResponse:
        """
        Send a notification to one of your configured notification channels.

        **Note:** This method requires you to be [logged in](#login) to LMK.

        <details><summary>Usage Example</summary>
        <p>

        ```python
        import lmk

        # Check that the client is logged in (optional)
        # If using for the first time on a new device, call lmk.login()
        assert lmk.logged_in()

        lmk.notify("Hello, world!")
        ```

        </p>
        </details>

        :param message: The content of the notification you want to send
        :type message: str
        :param content_type: The MIME type of the message you want to send; ``text/plain`` and
        ``text/markdown`` are supported. Defaults to ``text/markdown``
        :type content_type: str, optional
        :param notification_channels: A list of notification channel IDs or notification channel
        objects that you want to send the notification to. If ``None`` and ``notify = True`` (the
        default), this will be sent to the default notification channel for your account, which is
        the primary email address associated with your account by default. Defaults to ``None``
        :type notification_channels: List[str | NotificationChannelResponse], optional
        :param notify: ``True`` if you want to send a notification to one or more of your configured
        notification channels. If ``False``, this notification will only be visible via the LMK web app.
        Defaults to ``True``
        :type notify: bool, optional
        :param async_req: ``True`` if you want to send the request asynchronously, in which case this
        method will return a coroutine. Defaults to ``False``.
        :type async_req: bool, optional

        :return: The event object corresponding to the sent notification
        :rtype: EventResponse
        """
        api = EventApi(self.client)
        kws = {}
        if notify:
            notify_kws = {}
            if notification_channels is not None:
                channels = []
                for channel in notification_channels:
                    if isinstance(channel, NotificationChannelResponse):
                        channel = channel.notification_channel_id
                    channels.append(channel)
                notify_kws["channel_ids"] = channels
            elif self.default_channel:
                notify_kws["channel_ids"] = [self.default_channel]
            kws["notification_config"] = EventNotificationConfiguration(
                notify=True, **notify_kws
            )

        return pipeline(async_req)(
            lambda _: self._get_access_token(async_req),
            lambda access_token: api.post_event(
                EventRequest(message=message, contentType=content_type, **kws),  # type: ignore
                async_req=async_req,
                _headers={"Authorization": f"Bearer {access_token}"},
            ),
        )

    def create_session(
        self,
        name: str,
        state: Optional[
            Union[Dict[str, Any], ProcessSessionState, JupyterSessionState]
        ] = None,
        async_req: bool = False,
    ) -> SessionResponse:
        """
        Create an interactive session, which you can use to remotely monitor a process
        or Jupyter Notebook remotely via the LMK web app. You shouldn't have to use this
        method directly in normal usage, rather it will be invoed by

        **Note:** This method requires you to be [logged in](#login) to LMK.

        :param name: The name of the session. This will appear in the LMK app.
        :type name: str
        :param state: The initial state parameters for the session. The ``type``
        field is always required, but the rest of the required fields depend on what
        type of session it is. See the REST API documentation for more information.
        :type state: Dict[str, Any]
        :param async_req: ``True`` if you want to send the request asynchronously, in which case this
        method will return a coroutine. Defaults to ``False``.
        :type async_req: bool, optional

        :return: The session object corresponding to the created session.
        :rtype: SessionResponse
        """
        api = SessionApi(self.client)

        if isinstance(state, dict):
            typed_state = CreateSessionRequestState.from_dict(state)
        else:
            typed_state = CreateSessionRequestState(actual_instance=state)  # type: ignore

        return pipeline(async_req)(
            lambda _: self._get_access_token(async_req),
            lambda access_token: api.create_session(
                CreateSessionRequest(
                    name=name,
                    state=typed_state,
                ),
                async_req=async_req,
                _headers={"Authorization": f"Bearer {access_token}"},
            ),
        )

    def end_session(
        self,
        session_id: str,
        async_req: bool = False,
    ) -> None:
        """
        End an interactive session. After the session has been ended, its state cannot
        be updated any more. Only call this when completely finished with using a session.

        :param session_id: The ID of a previously created session that has not been ended
        yet
        :type session_id: str
        :param async_req: ``True`` if you want to send the request asynchronously, in which case this
        method will return a coroutine. Defaults to ``False``.
        :type async_req: bool, optional

        :return: This method does not return anything
        :rtype: None
        """
        api = SessionApi(self.client)

        return pipeline(async_req)(
            lambda _: self._get_access_token(async_req),
            lambda access_token: api.end_session(
                session_id, _headers={"Authorization": f"Bearer {access_token}"}
            ),
        )

    @contextlib.asynccontextmanager
    async def session_connect(
        self, session_id: str, read_only: bool = True
    ) -> AsyncGenerator[WebSocket, None]:
        """
        Connect via a web socket to an interactive session. This allows you to send state
        updates to the session via a web socket, and receive remote state updates initiated
        through the LMK web app or API calls from other clients.

        :param session_id: the ID of a previously created session that has not been ended
        yet.
        :type session_id: str
        :param read_only: Indicate whether to connect in "read only" mode. This means that
        updates cannot be sent via the web socket, only received. Defaults to ``True``
        :type read_only: bool, optional

        :return: An asynchronous context manager yielding a ``WebSocket`` object
        :rtype: AsyncContextManager[WebSocket]
        """
        access_token = await self._get_access_token_async()

        url = self.client.configuration.host + f"/v1/session/ws?token={access_token}"

        loop = asyncio.get_running_loop()

        async def on_connect(ws: WebSocket):
            LOGGER.debug("Session websocket connected for %s", session_id)
            await ws.send(
                {
                    "event": "connect",
                    "data": {"sessionId": session_id, "readOnly": read_only},
                }
            )
            LOGGER.debug("Sent connected message for %s", session_id)

        on_connect_cb = async_callback(on_connect, loop=loop)

        async with aiohttp.ClientSession(conn_timeout=10) as session:
            async with WebSocket(session, url, timeout=0.5, heartbeat=1) as ws:
                await on_connect(ws)

                ws_connected.connect(on_connect_cb, ws)
                try:
                    yield ws
                finally:
                    ws_connected.disconnect(on_connect_cb, ws)


DEFAULT_INSTANCE = None

DEFAULT_INSTANCE_LOCK = threading.RLock()


def get_instance() -> Instance:
    global DEFAULT_INSTANCE
    with DEFAULT_INSTANCE_LOCK:
        if DEFAULT_INSTANCE is None:
            DEFAULT_INSTANCE = Instance()
        return DEFAULT_INSTANCE


def set_instance(instance: Instance) -> None:
    global DEFAULT_INSTANCE
    with DEFAULT_INSTANCE_LOCK:
        if DEFAULT_INSTANCE is instance:
            return
        if DEFAULT_INSTANCE is not None:
            DEFAULT_INSTANCE.close()
        DEFAULT_INSTANCE, old_instance = instance, DEFAULT_INSTANCE
        default_instance_changed.send(
            set_instance,
            old_instance=old_instance,
            new_instance=instance,
        )



# coding: utf-8

"""
    LMK API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.0.1
    Generated by: https://openapi-generator.tech
"""


from __future__ import annotations
from inspect import getfullargspec
import pprint
import re  # noqa: F401
import json


from typing import Optional
from pydantic import BaseModel, Field, StrictStr, validator
from lmk.generated.models.event_notification_configuration import EventNotificationConfiguration

class EventRequest(BaseModel):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    session_id: Optional[StrictStr] = Field(None, alias="sessionId")
    message: StrictStr = ...
    content_type: Optional[StrictStr] = Field(None, alias="contentType")
    notification_config: Optional[EventNotificationConfiguration] = Field(None, alias="notificationConfig")
    __properties = ["sessionId", "message", "contentType", "notificationConfig"]

    @validator('content_type')
    def content_type_validate_enum(cls, v):
        if v is None:
            return v

        if v not in ('text/plain', 'text/markdown'):
            raise ValueError("must validate the enum values ('text/plain', 'text/markdown')")
        return v

    class Config:
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> EventRequest:
        """Create an instance of EventRequest from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        # override the default output from pydantic by calling `to_dict()` of notification_config
        if self.notification_config:
            _dict['notificationConfig'] = self.notification_config.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> EventRequest:
        """Create an instance of EventRequest from a dict"""
        if obj is None:
            return None

        if type(obj) is not dict:
            return EventRequest.parse_obj(obj)

        _obj = EventRequest.parse_obj({
            "session_id": obj.get("sessionId"),
            "message": obj.get("message"),
            "content_type": obj.get("contentType"),
            "notification_config": EventNotificationConfiguration.from_dict(obj.get("notificationConfig")) if obj.get("notificationConfig") is not None else None
        })
        return _obj




import ptrace.debugger
import signal
import subprocess
import sys

def debugger_example(pid):
    debugger = ptrace.debugger.PtraceDebugger()

    print("Attach the running process %s" % pid)
    process = debugger.addProcess(pid, False)
    # process is a PtraceProcess instance
    print("IP before: %#x" % process.getInstrPointer())

    print("Execute a single step")
    process.singleStep()
    # singleStep() gives back control to the process. We have to wait
    # until the process is trapped again to retrieve the control on the
    # process.
    process.waitSignals(signal.SIGTRAP)
    print("IP after: %#x" % process.getInstrPointer())

    process.detach()
    debugger.quit()

def main():
    args = [sys.executable, '-c', 'import time; time.sleep(60)']
    child_popen = subprocess.Popen(args)
    debugger_example(child_popen.pid)
    child_popen.kill()
    child_popen.wait()

if __name__ == "__main__":
    main()



import asyncio
import os
import logging
import warnings
from typing import TYPE_CHECKING

import aiohttp
from blinker import signal

from lmk.jupyter.notebook_info import find_server_and_session
from lmk.jupyter.utils import background_ctx
from lmk.utils.blinker import wait_for_signal

if TYPE_CHECKING:
    from lmk.jupyter.widget import LMKWidget


colab_support_enabled_changed = signal("colab-support-enabled-changed")


LOGGER = logging.getLogger(__name__)

COLAB_SUPPORT_ENABLED = False

CALLBACK_REGISTERED = False


def is_google_colab() -> bool:
    """
    Check if the current context appears to be within a colab notebook
    """
    return bool(os.getenv("COLAB_RELEASE_TAG"))


def colab_support_enabled() -> bool:
    """
    Determine if colab support is currently enabled. If in a colab
    environment it will be enabled by default
    """
    return COLAB_SUPPORT_ENABLED


def enable_google_colab_support(check_if_colab: bool = True) -> None:
    """
    Enable Google Colab support. By default only tries to do this if we
    detect we're running in a colab notebook.
    """
    global COLAB_SUPPORT_ENABLED, CALLBACK_REGISTERED

    from lmk.jupyter.widget import get_widget

    if COLAB_SUPPORT_ENABLED:
        return

    if check_if_colab and not is_google_colab():
        return

    try:
        from google.colab import output
    except ImportError:
        warnings.warn(
            "Unable to import the `google.colab.output` module, google colab support is not enabled.",
            RuntimeWarning,
        )
        return

    def widget_update():
        with background_ctx(LOGGER, "colab.enable_google_colab_support"):
            LOGGER.debug("Sending widget state for colab support")
            get_widget().send_state()

    if not CALLBACK_REGISTERED:
        output.register_callback("lmk.widget.sync", widget_update)
        CALLBACK_REGISTERED = True

    output.enable_custom_widget_manager()

    COLAB_SUPPORT_ENABLED = True
    colab_support_enabled_changed.send(None, old_value=False, new_value=True)


def disable_google_colab_support() -> None:
    """ """
    global COLAB_SUPPORT_ENABLED

    if not COLAB_SUPPORT_ENABLED:
        return

    from google.colab import output

    output.disable_custom_widget_manager()

    COLAB_SUPPORT_ENABLED = False
    colab_support_enabled_changed.send(None, old_value=True, new_value=False)


async def observe_google_colab_url(widget: "LMKWidget") -> None:
    """ """

    def get_url(file_id):
        return f"https://colab.research.google.com/drive/{file_id}"

    file_id = None
    while True:
        if not colab_support_enabled():
            await wait_for_signal(colab_support_enabled_changed)
            continue

        async with aiohttp.ClientSession() as client_session:
            try:
                _, session = await find_server_and_session(client_session)
            except Exception:
                LOGGER.exception("Error getting session")
            else:
                if session is None:
                    LOGGER.error("Unable to get session")
                else:
                    new_file_id = session["path"].split("=", 1)[1]
                    if new_file_id and new_file_id != file_id:
                        file_id = new_file_id
                        url = get_url(file_id)
                        if url != widget.url:
                            widget.url = url
            finally:
                await asyncio.sleep(60)



import asyncio
import os
import sys

from lmk.processmon.client import send_signal
from lmk.utils import setup_logging


async def main():
    setup_logging()

    job_id = sys.argv[1]
    signal = sys.argv[2]
    if signal.isdigit():
        signal = int(signal)

    job_dir = os.path.expanduser(f"~/.lmk/jobs/{job_id}")
    socket_path = os.path.join(job_dir, "daemon.sock")
    await send_signal(socket_path, signal)


if __name__ == "__main__":
    asyncio.run(main())



from lmk.jupyter.utils import (
    is_jupyter,
    run_javascript,
    display_widget as _ipython_display_,
)
from lmk.jupyter.hooks import _jupyter_labextension_paths, _jupyter_nbextension_paths

__all__ = [
    "is_jupyter",
    "run_javascript",
    "_ipython_display_",
    "_jupyter_labextension_paths",
    "_jupyter_nbextension_paths",
]

try:
    from lmk.jupyter.widget import get_widget, set_widget, IPythonMonitoringState
except ImportError:
    pass
else:
    from lmk.jupyter import methods
    from lmk.jupyter.colab import enable_google_colab_support
    from lmk.jupyter.magics import register_magics

    __all__ += ["get_widget", "set_widget", "IPythonMonitoringState"]
    __all__ += methods.__all__  # type: ignore

    def __getattr__(name: str):
        if name in globals():
            return globals()[name]
        if name in methods.__all__:
            return getattr(methods, name)
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")

    register_magics(raise_on_no_shell=False)
    enable_google_colab_support()



import lmk.patches

lmk.patches.patch()

from lmk import jupyter, methods
from lmk.constants import VERSION as __version__
from lmk.instance import get_instance, set_instance
from lmk.jupyter import (
    _jupyter_labextension_paths,
    _jupyter_nbextension_paths,
)

__all__ = [
    "__version__",
    "jupyter",
    "get_instance",
    "set_instance",
    "_jupyter_labextension_paths",
    "_jupyter_nbextension_paths",
]
__all__ += methods.__all__


def __getattr__(name: str):
    if name in globals():
        return globals()[name]
    if name in methods.__all__:
        return getattr(methods, name)
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



# coding: utf-8

"""
    LMK API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.0.1
    Generated by: https://openapi-generator.tech
"""


from __future__ import annotations
from inspect import getfullargspec
import pprint
import re  # noqa: F401
import json


from typing import Optional
from pydantic import BaseModel, Field, StrictStr, validator

class PartialProcessSessionState(BaseModel):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    type: Optional[StrictStr] = None
    hostname: Optional[StrictStr] = None
    command: Optional[StrictStr] = None
    pid: Optional[float] = None
    notify_on: Optional[StrictStr] = Field(None, alias="notifyOn")
    notify_channel: Optional[StrictStr] = Field(None, alias="notifyChannel")
    exit_code: Optional[float] = Field(None, alias="exitCode")
    __properties = ["type", "hostname", "command", "pid", "notifyOn", "notifyChannel", "exitCode"]

    @validator('type')
    def type_validate_enum(cls, v):
        if v is None:
            return v

        if v not in ('process'):
            raise ValueError("must validate the enum values ('process')")
        return v

    class Config:
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> PartialProcessSessionState:
        """Create an instance of PartialProcessSessionState from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> PartialProcessSessionState:
        """Create an instance of PartialProcessSessionState from a dict"""
        if obj is None:
            return None

        if type(obj) is not dict:
            return PartialProcessSessionState.parse_obj(obj)

        _obj = PartialProcessSessionState.parse_obj({
            "type": obj.get("type"),
            "hostname": obj.get("hostname"),
            "command": obj.get("command"),
            "pid": obj.get("pid"),
            "notify_on": obj.get("notifyOn"),
            "notify_channel": obj.get("notifyChannel"),
            "exit_code": obj.get("exitCode")
        })
        return _obj




import asyncio
from aiohttp import web


async def main():
    socket_path = "./tmp.sock"

    app =  web.Application()

    app.add_routes([
        web.get("/", lambda res: web.json_response({"ok": True}))
    ])

    runner = web.AppRunner(app)
    await runner.setup()

    site = web.UnixSite(runner, socket_path)
    await site.start()

    await asyncio.sleep(1)
    print("Shutting down")

    await runner.cleanup()
    print("After cleanup")


if __name__ == "__main__":
    asyncio.run(main())



# coding: utf-8

"""
    LMK API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.0.1
    Generated by: https://openapi-generator.tech
"""


from __future__ import annotations
from inspect import getfullargspec
import pprint
import re  # noqa: F401
import json


from typing import List, Optional
from pydantic import BaseModel, Field, StrictStr, validator
from lmk.generated.models.actor_response import ActorResponse
from lmk.generated.models.notification_channel_info import NotificationChannelInfo

class EventResponse(BaseModel):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    event_id: StrictStr = Field(..., alias="eventId")
    session_id: Optional[StrictStr] = Field(None, alias="sessionId")
    user_id: StrictStr = Field(..., alias="userId")
    actor: ActorResponse = ...
    message: StrictStr = ...
    content_type: StrictStr = Field(..., alias="contentType")
    channels: List[NotificationChannelInfo] = ...
    created_at: StrictStr = Field(..., alias="createdAt")
    __properties = ["eventId", "sessionId", "userId", "actor", "message", "contentType", "channels", "createdAt"]

    @validator('content_type')
    def content_type_validate_enum(cls, v):
        if v not in ('text/plain', 'text/markdown'):
            raise ValueError("must validate the enum values ('text/plain', 'text/markdown')")
        return v

    class Config:
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> EventResponse:
        """Create an instance of EventResponse from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        # override the default output from pydantic by calling `to_dict()` of actor
        if self.actor:
            _dict['actor'] = self.actor.to_dict()
        # override the default output from pydantic by calling `to_dict()` of each item in channels (list)
        _items = []
        if self.channels:
            for _item in self.channels:
                if _item:
                    _items.append(_item.to_dict())
            _dict['channels'] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> EventResponse:
        """Create an instance of EventResponse from a dict"""
        if obj is None:
            return None

        if type(obj) is not dict:
            return EventResponse.parse_obj(obj)

        _obj = EventResponse.parse_obj({
            "event_id": obj.get("eventId"),
            "session_id": obj.get("sessionId"),
            "user_id": obj.get("userId"),
            "actor": ActorResponse.from_dict(obj.get("actor")) if obj.get("actor") is not None else None,
            "message": obj.get("message"),
            "content_type": obj.get("contentType"),
            "channels": [NotificationChannelInfo.from_dict(_item) for _item in obj.get("channels")] if obj.get("channels") is not None else None,
            "created_at": obj.get("createdAt")
        })
        return _obj




import asyncio
import os
import signal

from lmk.process import exc
from lmk.process.attach import attach_simple
from lmk.process.client import wait_for_job
from lmk.process.daemon import ProcessMonitorController, ProcessMonitorDaemon, pid_ctx
from lmk.process.manager import JobManager
from lmk.process.monitor import ProcessMonitor
from lmk.utils.asyncio import asyncio_create_task, async_signal_handler_ctx
from lmk.utils.os import socket_exists


async def run_foreground(
    job_name: str,
    monitor: ProcessMonitor,
    manager: JobManager,
    log_level: str = "INFO",
    attach: bool = True,
) -> None:
    controller = ProcessMonitorController(job_name, monitor, manager)

    with pid_ctx(manager.pid_file(job_name), os.getpid()):
        log_path = manager.log_file(job_name)
        tasks = []
        tasks.append(asyncio_create_task(controller.run(log_path, log_level)))
        if attach:
            tasks.append(asyncio_create_task(attach_simple(job_name, manager)))

        async with async_signal_handler_ctx(
            [signal.SIGINT, signal.SIGTERM],
            lambda signum: controller.send_signal(signum),
        ):
            await asyncio.wait(tasks)


async def run_daemon(
    job_name: str, monitor: ProcessMonitor, manager: JobManager, log_level: str = "INFO"
) -> None:
    process = ProcessMonitorDaemon(job_name, monitor, manager.base_path, log_level)
    process.start()

    socket_path = manager.socket_file(job_name)

    async def wait():
        while not socket_exists(socket_path):
            job = await manager.get_job(job_name)
            if job.ended_at:
                break
            await asyncio.sleep(0.1)

    try:
        await asyncio.wait_for(wait(), 10)
    except asyncio.TimeoutError as err:
        raise Exception("Timed out waiting for monitoring process to come up") from err

    if socket_exists(socket_path):
        await wait_for_job(socket_path, "attach")
        return

    job = await manager.get_job(job_name)
    if job is None:
        raise exc.JobNotFound(job_name)

    raise exc.JobRaisedError(job)



# coding: utf-8

"""
    LMK API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.0.1
    Generated by: https://openapi-generator.tech
"""


from __future__ import annotations
from inspect import getfullargspec
import pprint
import re  # noqa: F401
import json



from pydantic import BaseModel, StrictStr
from lmk.generated.models.create_session_request_state import CreateSessionRequestState

class CreateSessionRequest(BaseModel):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    state: CreateSessionRequestState = ...
    name: StrictStr = ...
    __properties = ["state", "name"]

    class Config:
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> CreateSessionRequest:
        """Create an instance of CreateSessionRequest from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        # override the default output from pydantic by calling `to_dict()` of state
        if self.state:
            _dict['state'] = self.state.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> CreateSessionRequest:
        """Create an instance of CreateSessionRequest from a dict"""
        if obj is None:
            return None

        if type(obj) is not dict:
            return CreateSessionRequest.parse_obj(obj)

        _obj = CreateSessionRequest.parse_obj({
            "state": CreateSessionRequestState.from_dict(obj.get("state")) if obj.get("state") is not None else None,
            "name": obj.get("name")
        })
        return _obj




import abc
from typing import List


class MonitoredProcess(abc.ABC):
    pid: int

    command: List[str]

    @abc.abstractmethod
    async def wait(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    async def send_signal(self, signum: int) -> None:
        raise NotImplementedError


class ProcessMonitor(abc.ABC):
    """ """

    @abc.abstractmethod
    async def attach(
        self,
        output_path: str,
        log_path: str,
        log_level: str,
    ) -> MonitoredProcess:
        """ """
        raise NotImplementedError



# coding: utf-8

"""
    LMK API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.0.1
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import re  # noqa: F401

from pydantic import validate_arguments, ValidationError
from typing_extensions import Annotated

from lmk.generated.models.notification_channels_response import NotificationChannelsResponse

from lmk.generated.api_client import ApiClient
from lmk.generated.exceptions import (  # noqa: F401
    ApiTypeError,
    ApiValueError
)


class NotificationApi(object):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None):
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client

    @validate_arguments
    def list_notification_channels(self, **kwargs) -> NotificationChannelsResponse:  # noqa: E501
        """  # noqa: E501

        List the notification channels for the current user. If called by an oauth app, the results will depend on what scopes have been granted to the app. If `channel.read` has not been granted, this will return an empty list of channels. Otherwise, it will return all of the channels unless the user limited the app's visibillity during the oauth flow.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.list_notification_channels(async_req=True)
        >>> result = thread.get()

        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: NotificationChannelsResponse
        """
        kwargs['_return_http_data_only'] = True
        return self.list_notification_channels_with_http_info(**kwargs)  # noqa: E501

    @validate_arguments
    def list_notification_channels_with_http_info(self, **kwargs):  # noqa: E501
        """  # noqa: E501

        List the notification channels for the current user. If called by an oauth app, the results will depend on what scopes have been granted to the app. If `channel.read` has not been granted, this will return an empty list of channels. Otherwise, it will return all of the channels unless the user limited the app's visibillity during the oauth flow.  # noqa: E501
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.list_notification_channels_with_http_info(async_req=True)
        >>> result = thread.get()

        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: tuple(NotificationChannelsResponse, status_code(int), headers(HTTPHeaderDict))
        """

        _params = locals()

        _all_params = [
        ]
        _all_params.extend(
            [
                'async_req',
                '_return_http_data_only',
                '_preload_content',
                '_request_timeout',
                '_request_auth',
                '_content_type',
                '_headers'
            ]
        )

        # validate the arguments
        for _key, _val in _params['kwargs'].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method list_notification_channels" % _key
                )
            _params[_key] = _val
        del _params['kwargs']

        _collection_formats = {}

        # process the path parameters
        _path_params = {}

        # process the query parameters
        _query_params = []

        # process the header parameters
        _header_params = dict(_params.get('_headers', {}))

        # process the form parameters
        _form_params = []
        _files = {}

        # process the body parameter
        _body_params = None

        # set the HTTP header `Accept`
        _header_params['Accept'] = self.api_client.select_header_accept(
            ['application/json'])  # noqa: E501

        # authentication setting
        _auth_settings = []  # noqa: E501

        _response_types_map = {
            '200': "NotificationChannelsResponse",
        }

        return self.api_client.call_api(
            '/v1/notificationChannel', 'GET',
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get('async_req'),
            _return_http_data_only=_params.get('_return_http_data_only'),  # noqa: E501
            _preload_content=_params.get('_preload_content', True),
            _request_timeout=_params.get('_request_timeout'),
            collection_formats=_collection_formats,
            _request_auth=_params.get('_request_auth'))



import dataclasses as dc
import os
import re
from typing import Optional, Tuple, List

import click

from lmk.process import exc


PROJECT_DIR = os.path.dirname(__file__)

SCRIPT_PATH = os.path.join(PROJECT_DIR, "shell_cli.sh")


def get_shell_cli_script(shell: Optional[str] = None) -> str:
    with open(SCRIPT_PATH) as f:
        return f.read()


def detect_shell() -> Optional[str]:
    shell_var = os.getenv("SHELL")
    if shell_var is None:
        return None
    _, name = os.path.split(shell_var)
    return name


def shell_profile_file(shell: Optional[str] = None) -> str:
    if shell is None:
        raise exc.CannotDetermineShell()

    candidates = [f"~/.{shell}rc", f"~/.{shell}_profile"]
    for candidate in candidates:
        fullpath = os.path.expanduser(candidate)
        if os.path.exists(fullpath):
            return candidate

    raise RuntimeError(f"Unable to find shell profile file for shell: {shell}")


SCRIPT_START = "\n# ====== START LMK SHELL PLUGIN ======\n"

SCRIPT_END = "\n# ====== END LMK SHELL PLUGIN ======\n"


def install_script(profile_path: str) -> None:
    profile_path = os.path.expanduser(profile_path)

    with open(profile_path) as f:
        profile = f.read()

    start_index = -1
    try:
        start_index = profile.index(SCRIPT_START)
    except ValueError:
        pass

    end_index = -1
    if start_index != -1:
        try:
            end_index = profile.index(SCRIPT_END, start_index + len(SCRIPT_START))
        except ValueError:
            pass

    if start_index != -1 and end_index != -1:
        new_profile = "".join(
            [
                profile[:start_index],
                SCRIPT_START,
                f". {SCRIPT_PATH}",
                SCRIPT_END,
                profile[end_index + len(SCRIPT_END) :],
            ]
        )
    else:
        new_profile = "".join(
            [
                profile,
                SCRIPT_START,
                f". {SCRIPT_PATH}",
                SCRIPT_END,
            ]
        )

    with open(profile_path, "w+") as f:
        f.write(new_profile)


def uninstall_script(profile_path: str) -> None:
    profile_path = os.path.expanduser(profile_path)

    with open(profile_path) as f:
        profile = f.read()

    start_index = -1
    try:
        start_index = profile.index(SCRIPT_START)
    except ValueError:
        pass

    end_index = -1
    if start_index != -1:
        try:
            end_index = profile.index(SCRIPT_END, start_index + len(SCRIPT_START))
        except ValueError:
            pass

    new_profile = profile
    if start_index != -1 and end_index != -1:
        new_profile = "".join(
            [profile[:start_index], profile[end_index + len(SCRIPT_END) :]]
        )

    if new_profile != profile:
        with open(profile_path, "w+") as f:
            f.write(new_profile)


@dc.dataclass(frozen=True)
class ShellJob:
    command: str
    pid: int
    job_id: int
    state: str


def parse_jobs_output(output: str) -> List[ShellJob]:
    lines = output.strip().split("\n")
    pattern = re.compile(r"\[(\d+)\]\s+[\+\-]?\s+(\d+)\s+(\w+)\s+(.+)$")
    out = []

    for line in lines:
        match = pattern.match(line.strip())
        if not match:
            continue
        job_id = int(match.group(1))
        pid = int(match.group(2))
        state = match.group(3)
        command = match.group(4)
        out.append(ShellJob(command, pid, job_id, state))

    return out


def resolve_pid(pid: str) -> Tuple[int, Optional[int]]:
    if pid.isdigit():
        return int(pid), None

    if not pid.startswith("%"):
        click.secho(f"Invalid pid: {pid}", fg="red", bold=True)
        raise click.Abort

    job_id = int(pid[1:])

    shell_jobs = os.getenv("SHELL_JOBS")
    if shell_jobs is None:
        click.secho(
            "Cannot use bash job syntax without the shell plugin;"
            "install the shell plugin using `lmk shell-plugin --install`"
        )
        raise click.Abort

    jobs = parse_jobs_output(shell_jobs)
    match = None
    for job in jobs:
        if job.job_id == job_id:
            match = job
            break

    if match is None:
        raise exc.NoShellJobFound(job_id)

    return match.pid, match.job_id



from __future__ import absolute_import

# flake8: noqa

# import apis into api package
from lmk.generated.api.app_api import AppApi
from lmk.generated.api.event_api import EventApi
from lmk.generated.api.headless_auth_api import HeadlessAuthApi
from lmk.generated.api.notification_api import NotificationApi
from lmk.generated.api.oauth_api import OauthApi
from lmk.generated.api.session_api import SessionApi



from lmk.instance import Instance, get_instance


extra_names = ["channels"]


def _is_valid_name(name: str) -> bool:
    if name in extra_names:
        return True
    if name.startswith("_"):
        return False
    if not hasattr(Instance, name):
        return False
    if isinstance(getattr(Instance, name), property):
        return False
    return True


__all__ = [name for name in dir(Instance) + extra_names if _is_valid_name(name)]


def __getattr__(name: str):
    if not _is_valid_name(name):
        raise AttributeError(name)
    instance = get_instance()
    return getattr(instance, name)



from typing import Union, Any

import aiohttp

from lmk.utils.ws import WebSocket


async def send_signal(socket_path: str, signal: Union[str, int]) -> None:
    connector = aiohttp.UnixConnector(path=socket_path)
    async with aiohttp.ClientSession(connector=connector) as session:
        async with session.post(
            "http://daemon/signal", json={"signal": signal}
        ) as response:
            if response.status == 200:
                return
            raise ResponseError(response.status, await response.text())


async def update_job(socket_path: str) -> None:
    connector = aiohttp.UnixConnector(path=socket_path)
    async with aiohttp.ClientSession(connector=connector) as session:
        async with session.post("http://daemon/update") as response:
            if response.status == 200:
                return
            raise ResponseError(response.status, await response.text())


async def wait_for_job(socket_path: str, wait_for: str = "run") -> Any:
    connector = aiohttp.UnixConnector(path=socket_path)
    async with aiohttp.ClientSession(connector=connector) as session:
        async with WebSocket(session, f"http://daemon/wait?wait_for={wait_for}") as ws:
            async for message in ws:
                if not message["ok"]:
                    raise Exception(f"{message['error_type']}: {message['error']}")
                return message


class ResponseError(Exception):
    """ """

    def __init__(self, status: int, data: str) -> None:
        self.status = status
        self.data = data
        super().__init__(f"Request failed with status {status}: {data}")



import asyncio
import contextlib
import json
import logging
import multiprocessing
import os
import signal
import socket
import textwrap
from functools import wraps
from typing import Optional, Callable, Any, cast, Awaitable, AsyncGenerator

from aiohttp import web

from lmk.generated.models.event_response import EventResponse
from lmk.generated.models.session_response import SessionResponse
from lmk.generated.models.process_session_state import ProcessSessionState
from lmk.instance import get_instance
from lmk.process import exc
from lmk.process.manager import JobManager
from lmk.process.models import Job
from lmk.process.monitor import ProcessMonitor, MonitoredProcess
from lmk.utils import (
    setup_logging,
    socket_exists,
    read_last_lines,
    shlex_join,
    asyncio_event,
)
from lmk.utils.ws import WebSocket


LOGGER = logging.getLogger(__name__)


def route_handler(func: Callable) -> Callable:
    """ """

    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except Exception:
            LOGGER.exception("Error running route handler")
            return web.json_response({"message": "Internal server error"}, status=500)

    return wrapper


@contextlib.contextmanager
def pid_ctx(pid_file: str, pid: int):
    with open(pid_file, "w+") as f:
        f.write(str(pid))
    try:
        yield
    finally:
        if os.path.exists(pid_file):
            os.remove(pid_file)


class ProcessMonitorController:
    """ """

    def __init__(
        self,
        job_name: str,
        monitor: ProcessMonitor,
        manager: JobManager,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> None:
        self.job_name = job_name
        self.manager = manager
        self.monitor = monitor
        self.hostname = socket.gethostname()

        self.done_event = asyncio_event(loop=loop)
        self.attached_event = asyncio_event(loop=loop)
        self.update_event = asyncio_event(loop=loop)
        self.session: Optional[SessionResponse] = None
        self.process: Optional[MonitoredProcess] = None

    def _should_notify(self, job: Job) -> bool:
        if job.notify_on == "error":
            return job.exit_code != 0
        if job.notify_on == "stop":
            return True
        return False

    @route_handler
    async def _wait_websocket(self, request: web.Request) -> web.WebSocketResponse:
        wait_for = request.query.get("wait_for") or "run"
        ws = web.WebSocketResponse()
        await ws.prepare(request)

        try:
            attached_set = self.attached_event.is_set()
            if not attached_set:
                await self.attached_event.wait()

                job = await self.manager.get_job(self.job_name)
                if job is None:
                    raise exc.JobNotFound(self.job_name)

                if job.error is not None:
                    await ws.send_json(
                        {
                            "ok": False,
                            "stage": "attach",
                            "error_type": job.error_type,
                            "error": job.error,
                        }
                    )
                    await ws.close()
                    return ws

            if wait_for == "attach":
                await ws.send_json({"ok": True, "stage": "attach"})
                await ws.close()
                return ws

            await self.done_event.wait()
            job = await self.manager.get_job(self.job_name)

            if job is None:
                raise exc.JobNotFound(self.job_name)

            if job.error is not None:
                await ws.send_json(
                    {
                        "ok": False,
                        "stage": "run",
                        "error_type": job.error_type,
                        "error": job.error,
                    }
                )
                await ws.close()
                return ws

            await ws.send_json({"ok": True, "stage": "run", "exit_code": job.exit_code})
            await ws.close()
            return ws
        except ConnectionResetError:
            LOGGER.info("Web socket closed unexpectedly. wait_for: %s", wait_for)
            return ws

    @route_handler
    async def _handle_update(self, request: web.Request) -> web.Response:
        self.update_event.set()

        return web.json_response({"ok": True})

    @route_handler
    async def _send_signal(self, request: web.Request) -> web.Response:
        body = await request.json()
        signum = body["signal"]
        if isinstance(signum, str):
            signum = getattr(signal.Signals, signum).value

        await self.send_signal(signum)

        return web.json_response({"ok": True})

    async def _handle_session_action(self, action: str, body: Optional[Any]) -> None:
        if action == "sendSignal" and body is not None:
            await self.send_signal(body["signal"])
            return
        LOGGER.error("Invalid session action: %s, body: %s", action, body)

    @contextlib.asynccontextmanager
    async def _session_ctx(self) -> AsyncGenerator[None, None]:
        instance = get_instance()
        job = await self.manager.get_job(self.job_name)
        if job is None:
            raise exc.JobNotFound(self.job_name)

        self.session = await cast(
            Awaitable[SessionResponse],
            instance.create_session(
                self.job_name,
                ProcessSessionState(
                    type="process",
                    hostname=self.hostname,
                    command=shlex_join(json.loads(job.command))
                    if job.command
                    else "<unknown>",
                    pid=cast(float, job.pid),
                    notifyOn=job.notify_on,
                    notifyChannel=job.channel_id,
                    exitCode=None,
                ),
                async_req=True,
            ),
        )

        LOGGER.info("Created session: %s", self.session.session_id)
        await self.manager.update_job(self.job_name, session_id=self.session.session_id)

        ws: WebSocket
        async with instance.session_connect(self.session.session_id, False) as ws:
            LOGGER.debug("Connected to session: %s", self.session.session_id)

            async def handle_updates():
                while True:
                    update_task = asyncio.create_task(self.update_event.wait())
                    done_task = asyncio.create_task(self.done_event.wait())
                    await asyncio.wait(
                        [update_task, done_task], return_when=asyncio.FIRST_COMPLETED
                    )

                    job = await self.manager.get_job(self.job_name)

                    if update_task.done():
                        self.update_event.clear()
                        await ws.send(
                            {
                                "notifyOn": job.notify_on,
                                "notifyChannel": job.channel_id,
                            }
                        )

                    if done_task.done():
                        LOGGER.info(
                            "Sending session exit message; exit code %s", job.exit_code
                        )
                        await ws.send(
                            {
                                "notifyOn": job.notify_on,
                                "notifyChannel": job.channel_id,
                                "exitCode": job.exit_code,
                            }
                        )
                        LOGGER.info("Sent message")
                        await ws.close()
                        break

            async def handle_messages():
                async for message in ws:
                    LOGGER.debug("Web socket message: %s", message)
                    if not message["ok"]:
                        LOGGER.error("Error in websocket: %s", message)
                        continue
                    msg_type = message["message"]["type"]
                    if msg_type == "update":
                        await self.manager.update_job(
                            self.job_name,
                            notify_on=message["message"]["session"]["state"][
                                "notifyOn"
                            ],
                            channel_id=message["message"]["session"]["state"].get(
                                "notifyChannel"
                            ),
                        )
                    elif msg_type == "action":
                        try:
                            action = message["message"]["action"]
                            body = message["message"].get("body")
                            await self._handle_session_action(action, body)
                        except Exception as err:
                            LOGGER.exception(
                                "Error running action for message: %s", message
                            )
                    else:
                        LOGGER.warn("Unhandled ws message type: %s", msg_type)

            updates_task = asyncio.create_task(handle_updates())
            messages_task = asyncio.create_task(handle_messages())
            try:
                yield
            finally:
                LOGGER.debug("Waiting for session tasks")
                await asyncio.gather(updates_task, messages_task)
                LOGGER.debug("Session tasks are done")
                if self.session is not None:
                    await cast(
                        Awaitable[None],
                        instance.end_session(self.session.session_id, async_req=True),
                    )

    async def _run_server(self) -> None:
        socket_path = self.manager.socket_file(self.job_name)

        app = web.Application()

        app.add_routes(
            [
                web.get("/wait", self._wait_websocket),
                web.post("/signal", self._send_signal),
                web.post("/update", self._handle_update),
            ]
        )

        runner = web.AppRunner(app)
        await runner.setup()

        site = web.UnixSite(runner, socket_path)
        await site.start()

        while True:
            try:
                await asyncio.sleep(3600)
            except asyncio.CancelledError:
                await runner.cleanup()
                if socket_exists(socket_path):
                    os.remove(socket_path)
                raise

    async def _run_command(self, log_path: str, log_level: str) -> None:
        async with contextlib.AsyncExitStack() as stack:
            output_path = self.manager.output_file(self.job_name)
            job = await self.manager.get_job(self.job_name)

            try:
                if job is None:
                    raise exc.JobNotFound(self.job_name)

                # Create the output file
                with open(output_path, "wb+") as f:
                    pass

                LOGGER.debug("Attaching to process")
                self.process = await self.monitor.attach(
                    output_path,
                    log_path,
                    log_level,
                )
                LOGGER.debug(
                    "Attached to %d (%s)", self.process.pid, self.process.command
                )
                await self.manager.update_job(
                    self.job_name,
                    pid=self.process.pid,
                    command=self.process.command,
                )
                self.attached_event.set()

                await stack.enter_async_context(self._session_ctx())

                LOGGER.debug(
                    "Entered session context: %s",
                    self.session.session_id if self.session else "<unknown>",
                )

                exit_code = await self.process.wait()
                job = await self.manager.end_job(
                    self.job_name,
                    exit_code=exit_code,
                )
            except Exception as err:
                job = await self.manager.end_job(self.job_name, exit_code=-1, error=err)
                if not self.attached_event.is_set():
                    self.attached_event.set()

                LOGGER.exception(
                    "%d: %s monitor raised exception",
                    job.pid,
                    type(self.monitor).__name__,
                )
            else:
                LOGGER.info("%d: exited with code %d", job.pid, job.exit_code)
            finally:
                should_notify = self._should_notify(cast(Job, job))

            self.done_event.set()

            notify_status = "none"
            if not should_notify:
                LOGGER.info(
                    "Not sending notification. Current notify on: %s", job.notify_on
                )
            else:
                LOGGER.info(
                    "Sending notification to channel %s. Current notify on: %s",
                    job.channel_id or "default",
                    job.notify_on,
                )
                instance = get_instance()
                try:
                    message = textwrap.dedent(
                        f"""
                        Process exited with code **{exit_code}**:
                        ```bash
                        {shlex_join(json.loads(job.command)) if job.command else "<unknown>"}
                        ```
                        Process ran on `{self.hostname}`

                        Started: {job.started_at.isoformat() if job.started_at else "<unknown>"}
                        
                        Ended: {job.ended_at.isoformat() if job.ended_at else "<unknown>"}
                        """
                    ).strip()

                    logs = "\n".join(read_last_lines(output_path, 10, 10000))

                    if logs:
                        message += f"\n\nMost recent logs:\n```\n{logs}\n```"

                    await cast(
                        Awaitable[EventResponse],
                        instance.notify(
                            message,
                            notification_channels=(
                                None if job.channel_id is None else [job.channel_id]
                            ),
                            async_req=True,
                        ),
                    )
                    notify_status = "success"
                except Exception:
                    LOGGER.exception(
                        "Failed to send notification to channel %s.", job.channel_id
                    )
                    notify_status = "failed"

            await self.manager.update_job(self.job_name, notify_status=notify_status)

    async def send_signal(self, signum: int) -> None:
        if self.process is None:
            raise exc.ProcessNotAttached
        await self.process.send_signal(signum)

    async def run(self, log_path: str, log_level: str) -> None:
        await self.manager.start_job(self.job_name)

        tasks = []
        tasks.append(asyncio.create_task(self._run_server()))

        LOGGER.info("Running main process")
        try:
            await self._run_command(log_path, log_level)
        except Exception:
            LOGGER.exception("Error running command")
            raise
        finally:
            for task in reversed(tasks):
                task.cancel()

            with contextlib.suppress(asyncio.CancelledError):
                await asyncio.wait(tasks)


class ProcessMonitorDaemon(multiprocessing.Process):
    """ """

    def __init__(
        self,
        job_name: str,
        monitor: ProcessMonitor,
        base_path: Optional[str] = None,
        log_level: str = "INFO",
    ) -> None:
        super().__init__(daemon=False)
        self.job_name = job_name
        self.monitor = monitor
        self.base_path = base_path
        self.log_level = log_level

    def run(self) -> None:
        # Double fork so the process continues to run
        if os.fork() != 0:
            return

        # Detach this process from the parent so we don't share
        # signals
        os.setsid()

        manager = JobManager(self.base_path)
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        controller = ProcessMonitorController(
            self.job_name, self.monitor, manager, loop=loop
        )

        try:
            log_path = manager.log_file(self.job_name)
            pid_file = manager.pid_file(self.job_name)
            with open(log_path, "a+") as log_stream:
                setup_logging(log_stream=log_stream, level=self.log_level)

                with pid_ctx(pid_file, cast(int, self.pid)):
                    try:
                        loop.run_until_complete(
                            controller.run(log_path, self.log_level)
                        )
                    finally:
                        loop.run_until_complete(loop.shutdown_asyncgens())
                        loop.close()
        except:
            LOGGER.exception("Error running process monitor daemon")
            raise



import asyncio
from typing import Any

import blinker


def wait_for_signal(
    signal: blinker.Signal, sender: Any = blinker.ANY
) -> asyncio.Future:
    future: asyncio.Future = asyncio.Future()

    def handle_signal(sender, **kwargs):
        future.set_result((sender, kwargs))
        signal.disconnect(handle_signal, sender)

    signal.connect(handle_signal, sender)

    return future



import asyncio
import io
import logging
import signal
import sys
from typing import Optional, IO

from lmk.process import exc
from lmk.process.client import send_signal, wait_for_job
from lmk.process.manager import JobManager
from lmk.utils import wait_for_socket, shutdown_process, socket_exists, input_async


LOGGER = logging.getLogger(__name__)


class ProcessAttachment:
    def __init__(
        self, process: asyncio.subprocess.Process, job_name: str, manager: JobManager
    ):
        self.process = process
        self.job_name = job_name
        self.manager = manager

    def pause(self) -> None:
        self.process.send_signal(signal.SIGSTOP)

    def resume(self) -> None:
        self.process.send_signal(signal.SIGCONT)

    async def stop(self) -> None:
        await shutdown_process(self.process, 1, 1)

    async def wait(self) -> int:
        socket_path = self.manager.socket_file(self.job_name)
        if socket_exists(socket_path):
            resp = await wait_for_job(socket_path)
            exit_code = resp["exit_code"]
        else:
            job = await self.manager.get_job(self.job_name)
            if job is None:
                raise exc.JobNotFound(self.job_name)

            exit_code = job.exit_code
            if exit_code is None:
                raise exc.JobExitedUnexpectedly(job)

        LOGGER.info("Job %s exited with code %d", self.job_name, exit_code)
        await self.stop()
        return exit_code


async def attach(
    job_name: str,
    manager: JobManager,
    stdout_stream: IO[str] = sys.stdout,
    stderr_stream: IO[str] = sys.stderr,
) -> ProcessAttachment:
    log_file = manager.output_file(job_name)
    socket_path = manager.socket_file(job_name)

    await wait_for_socket(socket_path, 3)

    tail = await asyncio.create_subprocess_exec(
        "tail",
        "-f",
        log_file,
        stdin=asyncio.subprocess.DEVNULL,
        stdout=stdout_stream,
        stderr=stderr_stream,
        bufsize=0,
        start_new_session=True,
    )

    return ProcessAttachment(tail, job_name, manager)


async def attach_simple(
    job_name: str,
    manager: JobManager,
    stdout_stream: IO[str] = sys.stdout,
    stderr_stream: IO[str] = sys.stderr,
) -> int:
    attachment = await attach(job_name, manager, stdout_stream, stderr_stream)
    try:
        return await attachment.wait()
    except:
        await attachment.stop()
        raise


async def get_interrupt_action() -> str:
    while True:
        try:
            input_value = await input_async("interrupt/detach/resume process (i/d/r): ")
            input_value = input_value.lower().strip()
            if input_value in {"i", "d", "r"}:
                return input_value
        except KeyboardInterrupt:
            return "d"
        else:
            print(f"Invalid selection: {input_value}")


async def attach_interactive(job_name: str, manager: JobManager) -> Optional[int]:
    socket_path = manager.socket_file(job_name)

    attachment = await attach(job_name, manager)

    interupts = 0
    while True:
        task = asyncio.create_task(attachment.wait())
        try:
            return await asyncio.shield(task)
        except (asyncio.CancelledError, KeyboardInterrupt):
            if interupts > 0:
                task.cancel()
                await attachment.stop()
                break

            attachment.pause()

            action_task = asyncio.create_task(get_interrupt_action())

            await asyncio.wait([action_task, task], return_when=asyncio.FIRST_COMPLETED)

            if task.done():
                return task.result()

            task.cancel()

            action = action_task.result()
            if action == "i":
                attachment.resume()
                await send_signal(socket_path, signal.SIGINT)
                interupts += 1

            if action == "d":
                await attachment.stop()
                break

            if action == "r":
                attachment.resume()

    return None



# This was originally copied from the source code of the ipynbname package,
# didn't want to add a dependency for this relatively simple functionality
# source: https://github.com/msm1089/ipynbname/blob/master/ipynbname/__init__.py
import asyncio
import json
import logging
import os
from itertools import chain
from pathlib import Path
from typing import Generator, Tuple, Union, NoReturn, Optional, Dict, Any, List
from urllib.parse import urljoin

import aiohttp
import ipykernel
from blinker import signal
from jupyter_core.paths import jupyter_runtime_dir
from traitlets.config import MultipleInstanceError


LOGGER = logging.getLogger(__name__)

notebook_name_changed = signal("notebook-name-changed")

FILE_ERROR = "Can't identify the notebook name."


def _list_maybe_running_servers(runtime_dir=None) -> Generator[dict, None, None]:
    """Iterate over the server info files of running notebook servers."""
    if runtime_dir is None:
        runtime_dir = jupyter_runtime_dir()
    runtime_dir = Path(runtime_dir)

    if runtime_dir.is_dir():
        for file_name in chain(
            runtime_dir.glob("nbserver-*.json"),  # jupyter notebook (or lab 2)
            runtime_dir.glob("jpserver-*.json"),  # jupyterlab 3
        ):
            yield json.loads(file_name.read_bytes())


def _get_kernel_id() -> str:
    """Returns the kernel ID of the ipykernel."""
    connection_file = Path(ipykernel.get_connection_file()).stem
    kernel_id = connection_file.split("-", 1)[1]
    return kernel_id


def kernel_id() -> Optional[str]:
    try:
        return _get_kernel_id()
    except (MultipleInstanceError, RuntimeError):
        LOGGER.debug("Unable to get the current kernel ID", exc_info=True)
        return None


async def _get_sessions(
    srv: Dict[str, Any], session: aiohttp.ClientSession
) -> List[Dict[str, Any]]:
    """Given a server, returns sessions, or HTTPError if access is denied.
    NOTE: Works only when either there is no security or there is token
    based security. An HTTPError is raised if unable to connect to a
    server.
    """
    url = srv["url"]
    token = srv.get("token")
    sessions_url = urljoin(url, "/api/sessions")

    # If no token, just try to make a request once
    if not token:
        response = await session.get(sessions_url)
        response.raise_for_status()
        return await response.json()

    # If there is a token, we try two approaches: first use the token in a header,
    # then try using a query param.
    response = await session.get(f"{sessions_url}?token={token}")
    try:
        response.raise_for_status()
        return await response.json()
    except aiohttp.ClientResponseError:
        pass

    response = await session.get(
        sessions_url, headers={"Authorization": f"token {token}"}
    )
    response.raise_for_status()

    return await response.json()


async def find_server_and_session(
    session: aiohttp.ClientSession,
) -> (Union[Tuple[Dict[str, Any], Dict[str, Any]], Tuple[None, None]]):
    """
    Find the server and session for the current Jupyter kernel
    """
    try:
        kernel_id = _get_kernel_id()
    except (MultipleInstanceError, RuntimeError):
        LOGGER.debug("Unable to get the current kernel ID", exc_info=True)
        return None, None  # Could not determine
    for srv in _list_maybe_running_servers():
        try:
            sessions = await _get_sessions(srv, session)
            for sess in sessions:
                if sess["kernel"]["id"] == kernel_id:
                    return srv, sess
        except Exception:
            # There may be stale entries in the runtime directory, so this
            # is expected to happen sometimes
            LOGGER.debug("Error getting sessions for server %r", srv, exc_info=True)
    return None, None


async def notebook_name(client_session: aiohttp.ClientSession) -> str:
    """Returns the short name of the notebook w/o the .ipynb extension,
    or raises a FileNotFoundError exception if it cannot be determined.
    """
    from lmk.jupyter.colab import colab_support_enabled

    _, session = await find_server_and_session(client_session)
    if session is None:
        raise FileNotFoundError(FILE_ERROR)

    if colab_support_enabled():
        session_name = session.get("notebook", {}).get("name")
        if session_name:
            return session_name

    path = session.get("notebook", {}).get("path")
    if path is None:
        raise FileNotFoundError(FILE_ERROR)

    return os.path.basename(path)


class NotebookInfoWatcher:
    """
    Periodically query the jupyter server to see if the notebook name has changed
    """

    def __init__(self, poll_interval: float = 30.0) -> None:
        self.poll_interval = poll_interval
        self.notebook_name: Optional[str] = None

    async def refresh(self) -> None:
        async with aiohttp.ClientSession() as session:
            current_name = await notebook_name(session)
            if current_name == self.notebook_name:
                return
            self.notebook_name, old_name = current_name, self.notebook_name
            notebook_name_changed.send(self, old_value=old_name, new_value=current_name)

    async def main_loop(self) -> NoReturn:
        while True:
            try:
                await self.refresh()
            except Exception:
                LOGGER.warn("Error getting notebook name", exc_info=True)
            await asyncio.sleep(self.poll_interval)



# coding: utf-8

"""
    LMK API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.0.1
    Generated by: https://openapi-generator.tech
"""


from __future__ import annotations
from inspect import getfullargspec
import pprint
import re  # noqa: F401
import json



from pydantic import BaseModel, Field, StrictBool, StrictStr
from lmk.generated.models.actor_response import ActorResponse
from lmk.generated.models.notification_channel_response_payload import NotificationChannelResponsePayload

class NotificationChannelResponse(BaseModel):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    payload: NotificationChannelResponsePayload = ...
    notification_channel_id: StrictStr = Field(..., alias="notificationChannelId")
    name: StrictStr = ...
    order: float = ...
    is_default: StrictBool = Field(..., alias="isDefault")
    is_managed: StrictBool = Field(..., alias="isManaged")
    is_verified: StrictBool = Field(..., alias="isVerified")
    verification_required: StrictBool = Field(..., alias="verificationRequired")
    created_at: StrictStr = Field(..., alias="createdAt")
    created_by_actor: ActorResponse = Field(..., alias="createdByActor")
    last_updated_at: StrictStr = Field(..., alias="lastUpdatedAt")
    last_updated_by_actor: ActorResponse = Field(..., alias="lastUpdatedByActor")
    __properties = ["payload", "notificationChannelId", "name", "order", "isDefault", "isManaged", "isVerified", "verificationRequired", "createdAt", "createdByActor", "lastUpdatedAt", "lastUpdatedByActor"]

    class Config:
        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> NotificationChannelResponse:
        """Create an instance of NotificationChannelResponse from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True,
                          exclude={
                          },
                          exclude_none=True)
        # override the default output from pydantic by calling `to_dict()` of payload
        if self.payload:
            _dict['payload'] = self.payload.to_dict()
        # override the default output from pydantic by calling `to_dict()` of created_by_actor
        if self.created_by_actor:
            _dict['createdByActor'] = self.created_by_actor.to_dict()
        # override the default output from pydantic by calling `to_dict()` of last_updated_by_actor
        if self.last_updated_by_actor:
            _dict['lastUpdatedByActor'] = self.last_updated_by_actor.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> NotificationChannelResponse:
        """Create an instance of NotificationChannelResponse from a dict"""
        if obj is None:
            return None

        if type(obj) is not dict:
            return NotificationChannelResponse.parse_obj(obj)

        _obj = NotificationChannelResponse.parse_obj({
            "payload": NotificationChannelResponsePayload.from_dict(obj.get("payload")) if obj.get("payload") is not None else None,
            "notification_channel_id": obj.get("notificationChannelId"),
            "name": obj.get("name"),
            "order": obj.get("order"),
            "is_default": obj.get("isDefault"),
            "is_managed": obj.get("isManaged"),
            "is_verified": obj.get("isVerified"),
            "verification_required": obj.get("verificationRequired"),
            "created_at": obj.get("createdAt"),
            "created_by_actor": ActorResponse.from_dict(obj.get("createdByActor")) if obj.get("createdByActor") is not None else None,
            "last_updated_at": obj.get("lastUpdatedAt"),
            "last_updated_by_actor": ActorResponse.from_dict(obj.get("lastUpdatedByActor")) if obj.get("lastUpdatedByActor") is not None else None
        })
        return _obj




import asyncio
import os
import signal
import sys

from lmk.processmon.attach import attach_simple
from lmk.processmon.daemon import ProcessMonitorController, pid_ctx
from lmk.processmon.lldb_monitor import LLDBProcessMonitor
from lmk.processmon.manager import JobManager
from lmk.utils import setup_logging, asyncio_create_task, async_signal_handler_ctx


async def main() -> None:
    setup_logging()

    pid = int(sys.argv[1])

    base_path = os.path.expanduser("~/.lmk")
    manager = JobManager(base_path)

    job = manager.create_job("test")
    print("JOB ID", job.job_id)

    monitor = LLDBProcessMonitor()

    with pid_ctx(job.pid_file, os.getpid()):
        controller = ProcessMonitorController(
            pid,
            monitor,
            job.job_dir,
        )
        run_task = asyncio_create_task(controller.run("/dev/stdout"))
        attach_task = asyncio_create_task(attach_simple(job.job_dir))

        async with async_signal_handler_ctx(
            [signal.SIGINT, signal.SIGTERM],
            lambda signum: controller.send_signal(signum)
        ):
            await asyncio.wait([run_task, attach_task])


if __name__ == "__main__":
    asyncio.run(main())



import logging

from lmk.process import exc


def get_log_level(log_level: str) -> int:
    """
    """
    mapping = logging.getLevelNamesMapping()

    values = set(mapping.values())

    if log_level.isdigit():
        log_level_int = int(log_level)
        if log_level_int not in values:
            raise exc.InvalidLogLevel(log_level)
        return log_level_int

    out_log_level = log_level.strip().upper()
    if out_log_level not in mapping:
        raise exc.InvalidLogLevel(log_level)
    
    return mapping[out_log_level]
